"""Graphviz DOT renderer for lineage graphs.

Produces DOT format output for use with Graphviz tools.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from truthound.lineage.visualization.protocols import (
    IGraphRenderer,
    RenderFormat,
    RenderConfig,
)

if TYPE_CHECKING:
    from truthound.lineage.base import LineageGraph


class GraphvizRenderer(IGraphRenderer):
    """Renders lineage graphs to Graphviz DOT format.

    DOT format can be used with Graphviz tools to generate
    SVG, PNG, PDF and other image formats.

    Example:
        >>> renderer = GraphvizRenderer()
        >>> dot_output = renderer.render(graph)
        >>> # Save and render: dot -Tpng output.dot -o output.png
    """

    # Node shapes by type
    NODE_SHAPES = {
        "SOURCE": "cylinder",
        "TABLE": "box3d",
        "FILE": "note",
        "STREAM": "parallelogram",
        "TRANSFORMATION": "component",
        "VALIDATION": "diamond",
        "MODEL": "hexagon",
        "REPORT": "tab",
        "EXTERNAL": "oval",
        "VIRTUAL": "none",
    }

    # Edge styles by type
    EDGE_STYLES = {
        "DERIVED_FROM": "solid",
        "VALIDATED_BY": "dashed",
        "USED_BY": "dotted",
        "TRANSFORMED_TO": "solid",
        "JOINED_WITH": "bold",
        "AGGREGATED_TO": "solid",
        "FILTERED_TO": "dashed",
        "DEPENDS_ON": "dotted",
    }

    @property
    def format(self) -> RenderFormat:
        return RenderFormat.DOT

    def render(
        self,
        graph: "LineageGraph",
        config: RenderConfig | None = None,
    ) -> str:
        """Render graph to DOT format.

        Args:
            graph: Lineage graph to render
            config: Render configuration

        Returns:
            DOT format string
        """
        config = config or RenderConfig()
        return self._build_dot(graph, config)

    def render_subgraph(
        self,
        graph: "LineageGraph",
        root_node_id: str,
        direction: str = "downstream",
        max_depth: int = -1,
        config: RenderConfig | None = None,
    ) -> str:
        """Render subgraph from root node.

        Args:
            graph: Full lineage graph
            root_node_id: Root node ID
            direction: "upstream" or "downstream"
            max_depth: Maximum depth
            config: Render configuration

        Returns:
            DOT format string
        """
        config = config or RenderConfig()

        if direction == "upstream":
            subgraph = graph.get_upstream(root_node_id, max_depth)
        else:
            subgraph = graph.get_downstream(root_node_id, max_depth)

        config.highlight_nodes = config.highlight_nodes + [root_node_id]

        return self._build_dot(subgraph, config)

    def _build_dot(
        self,
        graph: "LineageGraph",
        config: RenderConfig,
    ) -> str:
        """Build DOT format string."""
        lines = []

        # Graph header
        lines.append("digraph LineageGraph {")
        lines.append("    // Graph settings")
        lines.append(f"    rankdir={config.orientation};")
        lines.append("    node [fontname=\"Arial\", fontsize=10];")
        lines.append("    edge [fontname=\"Arial\", fontsize=8];")
        lines.append("")

        # Node definitions
        lines.append("    // Nodes")
        node_ids = set()

        # Handle both list and dict node formats
        if isinstance(graph.nodes, dict):
            node_items = graph.nodes.items()
        else:
            node_items = [(node.id, node) for node in graph.nodes]

        for node_id, node in node_items:
            if config.filter_node_types and node.node_type.value not in config.filter_node_types:
                continue

            node_ids.add(node_id)

            shape = self.NODE_SHAPES.get(node.node_type.value, "ellipse")
            color = config.get_node_color(node.node_type.value)
            label = self._escape_label(node.name)

            # Build node attributes
            attrs = [
                f'label="{label}"',
                f'shape={shape}',
                f'fillcolor="{color}"',
                'style=filled',
            ]

            # Highlight
            if node_id in config.highlight_nodes:
                attrs.append('penwidth=3')
                attrs.append('color="red"')

            # Tooltip with metadata
            if config.include_metadata and hasattr(node, "metadata"):
                tooltip = f"{node.name}\\nType: {node.node_type.value}"
                if node.metadata.description:
                    tooltip += f"\\n{node.metadata.description}"
                attrs.append(f'tooltip="{self._escape_label(tooltip)}"')

            safe_id = self._safe_id(node_id)
            lines.append(f'    {safe_id} [{", ".join(attrs)}];')

        lines.append("")

        # Edge definitions
        lines.append("    // Edges")

        for edge in graph.edges:
            if config.filter_edge_types and edge.edge_type.value not in config.filter_edge_types:
                continue

            # Handle both source/source_id attribute names
            source_id = getattr(edge, 'source_id', None) or getattr(edge, 'source', None)
            target_id = getattr(edge, 'target_id', None) or getattr(edge, 'target', None)

            if source_id not in node_ids or target_id not in node_ids:
                continue

            style = self.EDGE_STYLES.get(edge.edge_type.value, "solid")
            color = config.get_edge_color(edge.edge_type.value)

            attrs = [
                f'style={style}',
                f'color="{color}"',
                f'label="{edge.edge_type.value}"',
            ]

            safe_source = self._safe_id(source_id)
            safe_target = self._safe_id(target_id)
            lines.append(f'    {safe_source} -> {safe_target} [{", ".join(attrs)}];')

        lines.append("}")

        return "\n".join(lines)

    def _escape_label(self, text: str) -> str:
        """Escape text for DOT labels."""
        return text.replace('"', '\\"').replace("\n", "\\n")

    def _safe_id(self, node_id: str) -> str:
        """Convert node ID to safe DOT identifier."""
        # Replace non-alphanumeric with underscore
        safe = "".join(c if c.isalnum() else "_" for c in node_id)
        # Ensure it starts with letter
        if safe and not safe[0].isalpha():
            safe = "n_" + safe
        return safe or "node"

    def render_svg(
        self,
        graph: "LineageGraph",
        config: RenderConfig | None = None,
    ) -> str:
        """Render graph to SVG using Graphviz.

        Requires graphviz package to be installed.

        Args:
            graph: Lineage graph
            config: Render configuration

        Returns:
            SVG string
        """
        try:
            import graphviz
        except ImportError:
            raise ImportError(
                "graphviz package is required for SVG rendering. "
                "Install with: pip install graphviz"
            )

        dot = self.render(graph, config)

        # Create graphviz source
        source = graphviz.Source(dot)
        svg = source.pipe(format="svg").decode("utf-8")

        return svg

    def render_png(
        self,
        graph: "LineageGraph",
        config: RenderConfig | None = None,
    ) -> bytes:
        """Render graph to PNG using Graphviz.

        Requires graphviz package to be installed.

        Args:
            graph: Lineage graph
            config: Render configuration

        Returns:
            PNG bytes
        """
        try:
            import graphviz
        except ImportError:
            raise ImportError(
                "graphviz package is required for PNG rendering. "
                "Install with: pip install graphviz"
            )

        dot = self.render(graph, config)

        source = graphviz.Source(dot)
        png = source.pipe(format="png")

        return png



"""Tests for escalation policy store implementations."""

from __future__ import annotations

from datetime import datetime, timedelta

import pytest

from truthound.checkpoint.escalation.protocols import EscalationRecord
from truthound.checkpoint.escalation.stores import (
    InMemoryEscalationStore,
    SQLiteEscalationStore,
    create_store,
)


class TestInMemoryEscalationStore:
    """Tests for InMemoryEscalationStore."""

    def test_save_and_get(self) -> None:
        """Test saving and retrieving a record."""
        store = InMemoryEscalationStore()
        record = EscalationRecord.create("inc-1", "policy-1")

        saved = store.save(record)
        retrieved = store.get(record.id)

        assert retrieved is not None
        assert retrieved.id == record.id
        assert retrieved.incident_id == "inc-1"

    def test_get_nonexistent(self) -> None:
        """Test getting nonexistent record."""
        store = InMemoryEscalationStore()
        result = store.get("nonexistent")
        assert result is None

    def test_get_by_incident(self) -> None:
        """Test getting records by incident ID."""
        store = InMemoryEscalationStore()

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record2 = EscalationRecord.create("inc-1", "policy-2")
        record3 = EscalationRecord.create("inc-2", "policy-1")

        store.save(record1)
        store.save(record2)
        store.save(record3)

        records = store.get_by_incident("inc-1")
        assert len(records) == 2
        assert all(r.incident_id == "inc-1" for r in records)

    def test_get_active(self) -> None:
        """Test getting active escalations."""
        store = InMemoryEscalationStore()

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "active"

        record2 = EscalationRecord.create("inc-2", "policy-1")
        record2.state = "resolved"

        record3 = EscalationRecord.create("inc-3", "policy-2")
        record3.state = "pending"

        store.save(record1)
        store.save(record2)
        store.save(record3)

        active = store.get_active()
        assert len(active) == 2

        active_policy1 = store.get_active("policy-1")
        assert len(active_policy1) == 1
        assert active_policy1[0].incident_id == "inc-1"

    def test_get_pending_escalations(self) -> None:
        """Test getting pending escalations."""
        store = InMemoryEscalationStore()
        now = datetime.now()

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "active"
        record1.next_escalation_at = now - timedelta(minutes=5)

        record2 = EscalationRecord.create("inc-2", "policy-1")
        record2.state = "active"
        record2.next_escalation_at = now + timedelta(minutes=5)

        store.save(record1)
        store.save(record2)

        pending = store.get_pending_escalations(now)
        assert len(pending) == 1
        assert pending[0].incident_id == "inc-1"

    def test_delete(self) -> None:
        """Test deleting a record."""
        store = InMemoryEscalationStore()
        record = EscalationRecord.create("inc-1", "policy-1")
        store.save(record)

        assert store.delete(record.id)
        assert store.get(record.id) is None
        assert not store.delete(record.id)  # Second delete returns False

    def test_cleanup_resolved(self) -> None:
        """Test cleaning up old resolved records."""
        store = InMemoryEscalationStore()

        # Old resolved record
        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "resolved"
        record1.resolved_at = datetime.now() - timedelta(days=2)
        store.save(record1)

        # Recent resolved record
        record2 = EscalationRecord.create("inc-2", "policy-1")
        record2.state = "resolved"
        record2.resolved_at = datetime.now()
        store.save(record2)

        # Active record
        record3 = EscalationRecord.create("inc-3", "policy-1")
        record3.state = "active"
        store.save(record3)

        removed = store.cleanup_resolved(timedelta(days=1))
        assert removed == 1
        assert store.get(record1.id) is None
        assert store.get(record2.id) is not None
        assert store.get(record3.id) is not None

    def test_get_stats(self) -> None:
        """Test getting store statistics."""
        store = InMemoryEscalationStore()

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "active"
        record1.current_level = 1
        store.save(record1)

        record2 = EscalationRecord.create("inc-2", "policy-1")
        record2.state = "acknowledged"
        record2.acknowledged_at = datetime.now()
        record2.current_level = 2
        store.save(record2)

        record3 = EscalationRecord.create("inc-3", "policy-2")
        record3.state = "resolved"
        record3.resolved_at = datetime.now()
        record3.current_level = 1
        store.save(record3)

        stats = store.get_stats()
        assert stats.total_escalations == 3
        assert stats.active_escalations == 1
        assert stats.acknowledged_count == 1
        assert stats.resolved_count == 1
        assert stats.escalations_by_level[1] == 2
        assert stats.escalations_by_level[2] == 1
        assert stats.escalations_by_policy["policy-1"] == 2
        assert stats.escalations_by_policy["policy-2"] == 1

    def test_clear(self) -> None:
        """Test clearing all records."""
        store = InMemoryEscalationStore()

        for i in range(5):
            record = EscalationRecord.create(f"inc-{i}", "policy-1")
            store.save(record)

        store.clear()
        assert len(store.get_active()) == 0


class TestSQLiteEscalationStore:
    """Tests for SQLiteEscalationStore."""

    def test_save_and_get(self) -> None:
        """Test saving and retrieving a record."""
        store = SQLiteEscalationStore(":memory:")
        record = EscalationRecord.create("inc-1", "policy-1")
        record.context = {"severity": "critical"}

        saved = store.save(record)
        retrieved = store.get(record.id)

        assert retrieved is not None
        assert retrieved.id == record.id
        assert retrieved.incident_id == "inc-1"
        assert retrieved.context["severity"] == "critical"

    def test_get_nonexistent(self) -> None:
        """Test getting nonexistent record."""
        store = SQLiteEscalationStore(":memory:")
        result = store.get("nonexistent")
        assert result is None

    def test_get_by_incident(self) -> None:
        """Test getting records by incident ID."""
        store = SQLiteEscalationStore(":memory:")

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record2 = EscalationRecord.create("inc-1", "policy-2")
        record3 = EscalationRecord.create("inc-2", "policy-1")

        store.save(record1)
        store.save(record2)
        store.save(record3)

        records = store.get_by_incident("inc-1")
        assert len(records) == 2

    def test_get_active(self) -> None:
        """Test getting active escalations."""
        store = SQLiteEscalationStore(":memory:")

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "active"
        store.save(record1)

        record2 = EscalationRecord.create("inc-2", "policy-1")
        record2.state = "resolved"
        store.save(record2)

        active = store.get_active()
        assert len(active) == 1
        assert active[0].incident_id == "inc-1"

    def test_get_pending_escalations(self) -> None:
        """Test getting pending escalations."""
        store = SQLiteEscalationStore(":memory:")
        now = datetime.now()

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "active"
        record1.next_escalation_at = now - timedelta(minutes=5)
        store.save(record1)

        record2 = EscalationRecord.create("inc-2", "policy-1")
        record2.state = "active"
        record2.next_escalation_at = now + timedelta(minutes=5)
        store.save(record2)

        pending = store.get_pending_escalations(now)
        assert len(pending) == 1
        assert pending[0].incident_id == "inc-1"

    def test_update_record(self) -> None:
        """Test updating an existing record."""
        store = SQLiteEscalationStore(":memory:")

        record = EscalationRecord.create("inc-1", "policy-1")
        record.state = "active"
        store.save(record)

        record.state = "acknowledged"
        record.acknowledged_at = datetime.now()
        record.acknowledged_by = "user-123"
        store.save(record)

        retrieved = store.get(record.id)
        assert retrieved is not None
        assert retrieved.state == "acknowledged"
        assert retrieved.acknowledged_by == "user-123"

    def test_delete(self) -> None:
        """Test deleting a record."""
        store = SQLiteEscalationStore(":memory:")
        record = EscalationRecord.create("inc-1", "policy-1")
        store.save(record)

        assert store.delete(record.id)
        assert store.get(record.id) is None

    def test_cleanup_resolved(self) -> None:
        """Test cleaning up old resolved records."""
        store = SQLiteEscalationStore(":memory:")

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "resolved"
        record1.resolved_at = datetime.now() - timedelta(days=2)
        store.save(record1)

        record2 = EscalationRecord.create("inc-2", "policy-1")
        record2.state = "active"
        store.save(record2)

        removed = store.cleanup_resolved(timedelta(days=1))
        assert removed == 1

    def test_get_stats(self) -> None:
        """Test getting store statistics."""
        store = SQLiteEscalationStore(":memory:")

        record1 = EscalationRecord.create("inc-1", "policy-1")
        record1.state = "active"
        store.save(record1)

        record2 = EscalationRecord.create("inc-2", "policy-2")
        record2.state = "resolved"
        record2.resolved_at = datetime.now()
        store.save(record2)

        stats = store.get_stats()
        assert stats.total_escalations == 2
        assert stats.active_escalations == 1
        assert stats.resolved_count == 1

    def test_clear(self) -> None:
        """Test clearing all records."""
        store = SQLiteEscalationStore(":memory:")

        for i in range(3):
            record = EscalationRecord.create(f"inc-{i}", "policy-1")
            store.save(record)

        store.clear()
        stats = store.get_stats()
        assert stats.total_escalations == 0

    def test_record_with_history(self) -> None:
        """Test saving record with history."""
        store = SQLiteEscalationStore(":memory:")

        record = EscalationRecord.create("inc-1", "policy-1")
        record.add_history_event("created", {"by": "system"})
        record.add_history_event("notified", {"target": "user-1"})
        store.save(record)

        retrieved = store.get(record.id)
        assert retrieved is not None
        assert len(retrieved.history) == 2
        assert retrieved.history[0]["event_type"] == "created"


class TestCreateStore:
    """Tests for store factory function."""

    def test_create_memory_store(self) -> None:
        """Test creating memory store."""
        store = create_store("memory")
        assert isinstance(store, InMemoryEscalationStore)

    def test_create_sqlite_store(self) -> None:
        """Test creating SQLite store."""
        store = create_store("sqlite", database=":memory:")
        assert isinstance(store, SQLiteEscalationStore)

    def test_create_unknown_store(self) -> None:
        """Test creating unknown store type."""
        with pytest.raises(ValueError, match="Unknown store type"):
            create_store("unknown")



"""MongoDB data source implementation.

This module provides async MongoDB data source with Motor driver support.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable
from urllib.parse import urlparse

from truthound.datasources._protocols import ColumnType, DataSourceCapability
from truthound.datasources.async_base import (
    AsyncConnectionPoolError,
    AsyncTimeoutError,
)
from truthound.datasources.nosql.base import (
    BaseNoSQLDataSource,
    NoSQLDataSourceConfig,
    NoSQLDataSourceError,
)

if TYPE_CHECKING:
    import polars as pl


# =============================================================================
# Exceptions
# =============================================================================


class MongoDBError(NoSQLDataSourceError):
    """MongoDB-specific error."""

    pass


class MongoDBConnectionError(MongoDBError):
    """MongoDB connection error."""

    def __init__(self, message: str, host: str | None = None) -> None:
        self.host = host
        super().__init__(f"MongoDB connection failed: {message}")


# =============================================================================
# Configuration
# =============================================================================


@dataclass
class MongoDBConfig(NoSQLDataSourceConfig):
    """Configuration for MongoDB data source.

    Supports connection via individual parameters or connection string.

    Attributes:
        host: MongoDB host.
        port: MongoDB port.
        database: Database name.
        collection: Collection name.
        username: Optional username.
        password: Optional password.
        auth_source: Authentication database.
        replica_set: Optional replica set name.
        tls: Whether to use TLS.
        tls_ca_file: Path to CA certificate file.
        tls_cert_key_file: Path to client certificate file.
        read_preference: Read preference (primary, primaryPreferred, etc.).
        write_concern: Write concern (0, 1, majority).
        app_name: Application name for connection.
    """

    host: str = "localhost"
    port: int = 27017
    database: str = ""
    collection: str = ""
    username: str | None = None
    password: str | None = None
    auth_source: str = "admin"
    replica_set: str | None = None
    tls: bool = False
    tls_ca_file: str | None = None
    tls_cert_key_file: str | None = None
    read_preference: str = "primary"
    write_concern: str = "1"
    app_name: str = "truthound"

    # Connection string (alternative to individual params)
    connection_string: str | None = None

    def get_connection_string(self) -> str:
        """Build MongoDB connection string.

        Returns:
            MongoDB URI string.
        """
        if self.connection_string:
            return self.connection_string

        # Build URI from parameters
        if self.username and self.password:
            auth = f"{self.username}:{self.password}@"
        else:
            auth = ""

        uri = f"mongodb://{auth}{self.host}:{self.port}"

        # Add options
        options = []
        if self.replica_set:
            options.append(f"replicaSet={self.replica_set}")
        if self.tls:
            options.append("tls=true")
        if self.auth_source and self.username:
            options.append(f"authSource={self.auth_source}")
        options.append(f"readPreference={self.read_preference}")
        options.append(f"appName={self.app_name}")

        if options:
            uri += "/?" + "&".join(options)

        return uri


# =============================================================================
# MongoDB Data Source
# =============================================================================


class MongoDBDataSource(BaseNoSQLDataSource):
    """Async MongoDB data source using Motor driver.

    Provides async access to MongoDB collections with automatic schema
    inference and Polars LazyFrame conversion.

    Example:
        >>> # From connection string
        >>> source = MongoDBDataSource.from_connection_string(
        ...     uri="mongodb://localhost:27017",
        ...     database="mydb",
        ...     collection="users",
        ... )
        >>>
        >>> async with source:
        ...     schema = await source.get_schema_async()
        ...     lf = await source.to_polars_lazyframe_async()
        ...     print(lf.collect())

        >>> # From config
        >>> config = MongoDBConfig(
        ...     host="localhost",
        ...     database="mydb",
        ...     collection="users",
        ... )
        >>> source = MongoDBDataSource(config)
    """

    source_type = "mongodb"

    def __init__(self, config: MongoDBConfig) -> None:
        """Initialize MongoDB data source.

        Args:
            config: MongoDB configuration.

        Raises:
            MongoDBError: If database or collection not specified.
        """
        if not config.database:
            raise MongoDBError("Database name is required")
        if not config.collection:
            raise MongoDBError("Collection name is required")

        super().__init__(config)
        self._client: Any = None
        self._collection: Any = None

    @property
    def config(self) -> MongoDBConfig:
        """Get typed configuration."""
        return self._config  # type: ignore

    @property
    def name(self) -> str:
        """Get data source name."""
        if self._config.name:
            return self._config.name
        return f"mongodb://{self.config.database}.{self.config.collection}"

    @property
    def database(self) -> str:
        """Get database name."""
        return self.config.database

    @property
    def collection_name(self) -> str:
        """Get collection name."""
        return self.config.collection

    @property
    def capabilities(self) -> set[DataSourceCapability]:
        """Get data source capabilities."""
        return {
            DataSourceCapability.SCHEMA_INFERENCE,
            DataSourceCapability.SAMPLING,
            DataSourceCapability.STREAMING,
            DataSourceCapability.ROW_COUNT,
        }

    # -------------------------------------------------------------------------
    # Factory Methods
    # -------------------------------------------------------------------------

    @classmethod
    def from_connection_string(
        cls,
        uri: str,
        database: str,
        collection: str,
        **kwargs: Any,
    ) -> "MongoDBDataSource":
        """Create data source from connection string.

        Args:
            uri: MongoDB connection URI.
            database: Database name.
            collection: Collection name.
            **kwargs: Additional configuration options.

        Returns:
            MongoDBDataSource instance.

        Example:
            >>> source = MongoDBDataSource.from_connection_string(
            ...     uri="mongodb://user:pass@host:27017",
            ...     database="mydb",
            ...     collection="users",
            ...     max_documents=10000,
            ... )
        """
        # Parse URI to extract host info
        parsed = urlparse(uri)
        host = parsed.hostname or "localhost"
        port = parsed.port or 27017

        config = MongoDBConfig(
            connection_string=uri,
            host=host,
            port=port,
            database=database,
            collection=collection,
            **kwargs,
        )
        return cls(config)

    @classmethod
    def from_atlas(
        cls,
        cluster_url: str,
        database: str,
        collection: str,
        username: str,
        password: str,
        **kwargs: Any,
    ) -> "MongoDBDataSource":
        """Create data source for MongoDB Atlas.

        Args:
            cluster_url: Atlas cluster URL (e.g., "cluster0.xxxxx.mongodb.net").
            database: Database name.
            collection: Collection name.
            username: Atlas username.
            password: Atlas password.
            **kwargs: Additional configuration options.

        Returns:
            MongoDBDataSource instance.

        Example:
            >>> source = MongoDBDataSource.from_atlas(
            ...     cluster_url="cluster0.xxxxx.mongodb.net",
            ...     database="mydb",
            ...     collection="users",
            ...     username="user",
            ...     password="pass",
            ... )
        """
        uri = (
            f"mongodb+srv://{username}:{password}@{cluster_url}"
            f"/?retryWrites=true&w=majority"
        )
        config = MongoDBConfig(
            connection_string=uri,
            database=database,
            collection=collection,
            tls=True,
            **kwargs,
        )
        return cls(config)

    # -------------------------------------------------------------------------
    # Connection Management
    # -------------------------------------------------------------------------

    async def _create_connection_factory(self) -> Callable:
        """Create a connection factory for the pool."""

        async def factory():
            try:
                from motor.motor_asyncio import AsyncIOMotorClient
            except ImportError:
                raise ImportError(
                    "Motor is required for MongoDB support. "
                    "Install it with: pip install motor"
                )

            client = AsyncIOMotorClient(
                self.config.get_connection_string(),
                maxPoolSize=self.config.pool_size,
                serverSelectionTimeoutMS=int(self.config.connection_timeout * 1000),
            )
            # Test connection
            await client.admin.command("ping")
            return client

        return factory

    async def connect_async(self) -> None:
        """Establish connection to MongoDB."""
        if self._is_connected:
            return

        async with self._lock:
            if self._is_connected:
                return

            try:
                from motor.motor_asyncio import AsyncIOMotorClient
            except ImportError:
                raise ImportError(
                    "Motor is required for MongoDB support. "
                    "Install it with: pip install motor"
                )

            try:
                self._client = AsyncIOMotorClient(
                    self.config.get_connection_string(),
                    maxPoolSize=self.config.pool_size,
                    serverSelectionTimeoutMS=int(
                        self.config.connection_timeout * 1000
                    ),
                )

                # Test connection
                await self._client.admin.command("ping")

                # Get database and collection
                db = self._client[self.config.database]
                self._collection = db[self.config.collection]

                # Pre-fetch schema
                self._cached_schema = await self.get_schema_async()

                self._is_connected = True

            except Exception as e:
                raise MongoDBConnectionError(
                    str(e), host=self.config.host
                )

    async def disconnect_async(self) -> None:
        """Close MongoDB connection."""
        if not self._is_connected:
            return

        async with self._lock:
            if not self._is_connected:
                return

            if self._client:
                self._client.close()
                self._client = None
                self._collection = None

            self._is_connected = False

    async def validate_connection_async(self) -> bool:
        """Validate MongoDB connection.

        Returns:
            True if connection is healthy.
        """
        try:
            if not self._is_connected:
                await self.connect_async()

            await self._client.admin.command("ping")
            return True
        except Exception:
            return False

    # -------------------------------------------------------------------------
    # Document Operations
    # -------------------------------------------------------------------------

    async def _fetch_sample_documents(
        self, n: int
    ) -> list[dict[str, Any]]:
        """Fetch sample documents for schema inference.

        Uses $sample aggregation for random sampling.

        Args:
            n: Number of documents to sample.

        Returns:
            List of sample documents.
        """
        if not self._is_connected:
            await self.connect_async()

        pipeline = [{"$sample": {"size": n}}]
        cursor = self._collection.aggregate(pipeline)
        documents = await cursor.to_list(length=n)

        # Convert ObjectId to string for JSON compatibility
        return [self._normalize_document(doc) for doc in documents]

    async def _fetch_documents(
        self,
        filter: dict[str, Any] | None = None,
        limit: int | None = None,
        skip: int | None = None,
    ) -> list[dict[str, Any]]:
        """Fetch documents from collection.

        Args:
            filter: MongoDB query filter.
            limit: Maximum documents to return.
            skip: Number of documents to skip.

        Returns:
            List of documents.
        """
        if not self._is_connected:
            await self.connect_async()

        cursor = self._collection.find(filter or {})

        if skip:
            cursor = cursor.skip(skip)
        if limit:
            cursor = cursor.limit(limit)

        documents = await cursor.to_list(length=limit or self.config.max_documents)
        return [self._normalize_document(doc) for doc in documents]

    async def _count_documents(
        self, filter: dict[str, Any] | None = None
    ) -> int:
        """Count documents in collection.

        Args:
            filter: Optional filter criteria.

        Returns:
            Document count.
        """
        if not self._is_connected:
            await self.connect_async()

        return await self._collection.count_documents(filter or {})

    def _normalize_document(self, doc: dict[str, Any]) -> dict[str, Any]:
        """Normalize document for Polars compatibility.

        Converts ObjectId, Binary, etc. to Python native types.

        Args:
            doc: MongoDB document.

        Returns:
            Normalized document.
        """
        normalized = {}

        for key, value in doc.items():
            # Skip internal MongoDB fields if needed
            if key == "_id":
                normalized[key] = str(value)
            elif hasattr(value, "__class__"):
                type_name = value.__class__.__name__
                if type_name == "ObjectId":
                    normalized[key] = str(value)
                elif type_name == "Binary":
                    normalized[key] = bytes(value)
                elif type_name == "Decimal128":
                    normalized[key] = float(value.to_decimal())
                elif type_name == "datetime":
                    normalized[key] = value
                elif isinstance(value, dict):
                    normalized[key] = self._normalize_document(value)
                elif isinstance(value, list):
                    normalized[key] = [
                        self._normalize_document(v) if isinstance(v, dict) else v
                        for v in value
                    ]
                else:
                    normalized[key] = value
            else:
                normalized[key] = value

        return normalized

    # -------------------------------------------------------------------------
    # Aggregation
    # -------------------------------------------------------------------------

    async def aggregate_async(
        self,
        pipeline: list[dict[str, Any]],
        allow_disk_use: bool = True,
    ) -> list[dict[str, Any]]:
        """Execute aggregation pipeline.

        Args:
            pipeline: MongoDB aggregation pipeline stages.
            allow_disk_use: Allow disk use for large operations.

        Returns:
            Aggregation results.

        Example:
            >>> pipeline = [
            ...     {"$match": {"status": "active"}},
            ...     {"$group": {"_id": "$category", "count": {"$sum": 1}}},
            ... ]
            >>> results = await source.aggregate_async(pipeline)
        """
        if not self._is_connected:
            await self.connect_async()

        cursor = self._collection.aggregate(
            pipeline,
            allowDiskUse=allow_disk_use,
        )
        documents = await cursor.to_list(length=self.config.max_documents)
        return [self._normalize_document(doc) for doc in documents]

    # -------------------------------------------------------------------------
    # Sampling
    # -------------------------------------------------------------------------

    async def sample_async(
        self, n: int = 1000, seed: int | None = None
    ) -> "MongoDBDataSource":
        """Create a sampled data source.

        Note: MongoDB's $sample doesn't support seeds, so results
        may vary between calls.

        Args:
            n: Number of documents to sample.
            seed: Ignored (MongoDB $sample doesn't support seeds).

        Returns:
            New MongoDBDataSource with sampled configuration.
        """
        # Create new config with reduced max_documents
        config = MongoDBConfig(
            connection_string=self.config.connection_string,
            host=self.config.host,
            port=self.config.port,
            database=self.config.database,
            collection=self.config.collection,
            username=self.config.username,
            password=self.config.password,
            name=f"{self.name}_sample",
            max_documents=n,
            schema_sample_size=min(n, self.config.schema_sample_size),
        )
        return MongoDBDataSource(config)

    # -------------------------------------------------------------------------
    # Additional Query Methods
    # -------------------------------------------------------------------------

    async def find_one_async(
        self,
        filter: dict[str, Any] | None = None,
    ) -> dict[str, Any] | None:
        """Find a single document.

        Args:
            filter: Query filter.

        Returns:
            Document or None if not found.
        """
        if not self._is_connected:
            await self.connect_async()

        doc = await self._collection.find_one(filter or {})
        return self._normalize_document(doc) if doc else None

    async def distinct_async(
        self,
        field: str,
        filter: dict[str, Any] | None = None,
    ) -> list[Any]:
        """Get distinct values for a field.

        Args:
            field: Field name.
            filter: Optional filter.

        Returns:
            List of distinct values.
        """
        if not self._is_connected:
            await self.connect_async()

        return await self._collection.distinct(field, filter or {})

    async def get_indexes_async(self) -> list[dict[str, Any]]:
        """Get collection indexes.

        Returns:
            List of index definitions.
        """
        if not self._is_connected:
            await self.connect_async()

        cursor = self._collection.list_indexes()
        return await cursor.to_list(length=100)



"""Tests for ObservableStore wrapper."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any
import uuid

import pytest

from truthound.stores.base import BaseStore, StoreConfig, StoreQuery, StoreNotFoundError
from truthound.stores.observability.audit import AuditEventType, InMemoryAuditBackend
from truthound.stores.observability.config import ObservabilityConfig, AuditConfig, MetricsConfig, TracingConfig
from truthound.stores.observability.metrics import InMemoryMetricsBackend
from truthound.stores.observability.protocols import ObservabilityContext
from truthound.stores.observability.store import ObservableStore, ObservabilityManager
from truthound.stores.observability.tracing import InMemoryTracer


# Mock store for testing
@dataclass
class MockItem:
    """Mock item for testing."""

    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    data_asset: str = "test-asset"
    run_time: datetime = field(default_factory=datetime.utcnow)

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "data_asset": self.data_asset,
            "run_time": self.run_time.isoformat(),
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "MockItem":
        return cls(
            id=data["id"],
            name=data.get("name", ""),
            data_asset=data.get("data_asset", "test-asset"),
        )


class MockStore(BaseStore[MockItem, StoreConfig]):
    """Mock store for testing."""

    def __init__(self) -> None:
        super().__init__()
        self._items: dict[str, MockItem] = {}

    @classmethod
    def _default_config(cls) -> StoreConfig:
        return StoreConfig()

    def _do_initialize(self) -> None:
        pass

    def save(self, item: MockItem) -> str:
        self._items[item.id] = item
        return item.id

    def get(self, item_id: str) -> MockItem:
        if item_id not in self._items:
            raise StoreNotFoundError("MockItem", item_id)
        return self._items[item_id]

    def exists(self, item_id: str) -> bool:
        return item_id in self._items

    def delete(self, item_id: str) -> bool:
        if item_id in self._items:
            del self._items[item_id]
            return True
        return False

    def list_ids(self, query: StoreQuery | None = None) -> list[str]:
        return list(self._items.keys())

    def query(self, query: StoreQuery) -> list[MockItem]:
        results = list(self._items.values())
        if query.data_asset:
            results = [r for r in results if r.data_asset == query.data_asset]
        if query.limit:
            results = results[: query.limit]
        return results


class TestObservabilityManager:
    """Tests for ObservabilityManager."""

    def test_create_with_defaults(self) -> None:
        config = ObservabilityConfig()
        manager = ObservabilityManager(config, store_type="test")

        assert manager.store_type == "test"
        assert manager.store_id is not None

    def test_audit_disabled(self) -> None:
        config = ObservabilityConfig(audit=AuditConfig(enabled=False))
        manager = ObservabilityManager(config, store_type="test")

        assert manager.audit is None

    def test_metrics_disabled(self) -> None:
        config = ObservabilityConfig(metrics=MetricsConfig(enabled=False))
        manager = ObservabilityManager(config, store_type="test")

        assert manager.metrics is None

    def test_create_context(self) -> None:
        config = ObservabilityConfig()
        manager = ObservabilityManager(config, store_type="test")

        ctx = manager.create_context(user_id="user-1")

        assert ctx.correlation_id is not None
        assert ctx.user_id == "user-1"

    def test_observe_success(self) -> None:
        audit_backend = InMemoryAuditBackend()
        metrics_backend = InMemoryMetricsBackend()
        tracer = InMemoryTracer()

        config = ObservabilityConfig()
        manager = ObservabilityManager(
            config,
            store_type="test",
            audit_backend=audit_backend,
            metrics_backend=metrics_backend,
            tracer=tracer,
        )

        with manager.observe(
            "save",
            AuditEventType.CREATE,
            resource_id="item-1",
        ):
            pass

        # Check audit
        assert len(audit_backend.events) == 1
        assert audit_backend.events[0].operation == "save"

        # Check metrics
        export = metrics_backend.export()
        assert "operations_total" in export

        # Check tracing
        assert len(tracer.spans) == 1
        assert tracer.spans[0].name == "store.save"

    def test_observe_failure(self) -> None:
        audit_backend = InMemoryAuditBackend()
        metrics_backend = InMemoryMetricsBackend()
        tracer = InMemoryTracer()

        config = ObservabilityConfig()
        manager = ObservabilityManager(
            config,
            store_type="test",
            audit_backend=audit_backend,
            metrics_backend=metrics_backend,
            tracer=tracer,
        )

        with pytest.raises(ValueError):
            with manager.observe(
                "save",
                AuditEventType.CREATE,
                resource_id="item-1",
            ):
                raise ValueError("Test error")

        # Check audit recorded failure
        assert len(audit_backend.events) == 1
        assert audit_backend.events[0].error_type == "ValueError"

        # Check metrics recorded error
        export = metrics_backend.export()
        assert "errors_total" in export


class TestObservableStore:
    """Tests for ObservableStore wrapper."""

    def test_save_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        item = MockItem(name="test-item")
        item_id = store.save(item)

        assert item_id == item.id
        assert base_store.exists(item_id)

    def test_get_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        # Save an item first
        item = MockItem(name="test")
        base_store.save(item)

        # Get via observable store
        result = store.get(item.id)
        assert result.name == "test"

    def test_delete_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        item = MockItem(name="test")
        base_store.save(item)

        result = store.delete(item.id)
        assert result is True
        assert not base_store.exists(item.id)

    def test_exists_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        item = MockItem()
        base_store.save(item)

        assert store.exists(item.id) is True
        assert store.exists("nonexistent") is False

    def test_query_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        for i in range(5):
            base_store.save(MockItem(name=f"item-{i}"))

        results = store.query(StoreQuery(limit=3))
        assert len(results) == 3

    def test_list_ids_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        for i in range(3):
            base_store.save(MockItem(name=f"item-{i}"))

        ids = store.list_ids()
        assert len(ids) == 3

    def test_count_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        for i in range(5):
            base_store.save(MockItem())

        assert store.count() == 5

    def test_clear_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        for i in range(5):
            base_store.save(MockItem())

        deleted = store.clear()
        assert deleted == 5
        assert store.count() == 0

    def test_save_batch(self) -> None:
        base_store = MockStore()
        audit_backend = InMemoryAuditBackend()
        config = ObservabilityConfig()

        store = ObservableStore(base_store, config)
        store._obs._audit_backend = audit_backend

        items = [MockItem(name=f"item-{i}") for i in range(10)]
        ids = store.save_batch(items)

        assert len(ids) == 10
        for item_id in ids:
            assert base_store.exists(item_id)

    def test_delete_batch(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        items = [MockItem() for _ in range(5)]
        for item in items:
            base_store.save(item)

        ids = [item.id for item in items[:3]]
        deleted = store.delete_batch(ids)

        assert deleted == 3
        assert store.count() == 2

    def test_with_context(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        context = ObservabilityContext(
            correlation_id="test-corr",
            user_id="user-1",
        )
        contextualized = store.with_context(context)

        assert contextualized._default_context == context
        assert contextualized._store is base_store

    def test_with_user(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        user_store = store.with_user("user-123", "tenant-456")

        assert user_store._default_context.user_id == "user-123"
        assert user_store._default_context.tenant_id == "tenant-456"

    def test_get_metrics(self) -> None:
        base_store = MockStore()
        metrics_backend = InMemoryMetricsBackend()
        config = ObservabilityConfig()

        store = ObservableStore(base_store, config)
        store._obs._metrics_backend = metrics_backend

        # Perform some operations
        item = MockItem()
        store.save(item)
        store.get(item.id)

        metrics = store.get_metrics()
        assert isinstance(metrics, str)

    def test_get_audit_events(self) -> None:
        base_store = MockStore()
        audit_backend = InMemoryAuditBackend()
        config = ObservabilityConfig()

        store = ObservableStore(base_store, config)
        store._obs._audit_backend = audit_backend

        # Perform some operations
        item = MockItem()
        store.save(item)

        events = store.get_audit_events()
        # Events may or may not be recorded depending on backend setup
        assert isinstance(events, list)

    def test_flush_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        # Should not raise
        store.flush_observability()

    def test_close_with_observability(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        # Should not raise
        store.close()

    def test_error_propagation(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        with pytest.raises(StoreNotFoundError):
            store.get("nonexistent-id")

    def test_config_access(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig.disabled()
        store = ObservableStore(base_store, config)

        assert store.config == base_store.config

    def test_observability_manager_access(self) -> None:
        base_store = MockStore()
        config = ObservabilityConfig()
        store = ObservableStore(base_store, config)

        assert store.observability is not None
        assert store.observability.store_type == "MockStore"


class TestObservabilityConfig:
    """Tests for ObservabilityConfig factory methods."""

    def test_disabled_config(self) -> None:
        config = ObservabilityConfig.disabled()

        assert config.audit.enabled is False
        assert config.metrics.enabled is False
        assert config.tracing.enabled is False

    def test_minimal_config(self) -> None:
        config = ObservabilityConfig.minimal()

        assert config.audit.enabled is True
        assert config.metrics.enabled is True
        assert config.tracing.enabled is False

    def test_production_config(self) -> None:
        config = ObservabilityConfig.production("my-service")

        assert config.audit.enabled is True
        assert config.metrics.enabled is True
        assert config.tracing.enabled is True
        assert config.tracing.service_name == "my-service"
        assert config.environment == "production"



"""Trace context management for distributed tracing.

This module provides context propagation for distributed tracing,
enabling correlation of logs and metrics across service boundaries.

Features:
    - Trace and span context management
    - W3C Trace Context compatible headers
    - Automatic context propagation
    - Integration with logging and metrics

Architecture:
    TraceContext -> SpanContext -> (nested spans)

    Each trace has a unique trace_id and contains multiple spans.
    Spans can be nested to represent call hierarchies.
"""

from __future__ import annotations

import threading
import time
import uuid
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Callable, Iterator, TypeVar

F = TypeVar("F", bound=Callable[..., Any])


# =============================================================================
# Trace IDs
# =============================================================================


def generate_trace_id() -> str:
    """Generate a unique trace ID (128-bit hex string)."""
    return uuid.uuid4().hex


def generate_span_id() -> str:
    """Generate a unique span ID (64-bit hex string)."""
    return uuid.uuid4().hex[:16]


# =============================================================================
# Span Status
# =============================================================================


class SpanStatus(Enum):
    """Status of a span."""

    UNSET = "unset"
    OK = "ok"
    ERROR = "error"


# =============================================================================
# Span Context
# =============================================================================


@dataclass
class SpanContext:
    """Context for a single span within a trace.

    A span represents a single operation within a trace.
    Spans can be nested to form a tree structure.

    Attributes:
        trace_id: Unique trace identifier.
        span_id: Unique span identifier.
        parent_span_id: Parent span ID (None for root span).
        name: Human-readable span name.
        start_time: When the span started.
        end_time: When the span ended (None if active).
        status: Span status.
        attributes: Key-value attributes.
        events: List of events that occurred during span.
    """

    trace_id: str
    span_id: str
    parent_span_id: str | None = None
    name: str = ""
    start_time: datetime = field(default_factory=lambda: datetime.now(timezone.utc))
    end_time: datetime | None = None
    status: SpanStatus = SpanStatus.UNSET
    attributes: dict[str, Any] = field(default_factory=dict)
    events: list[dict[str, Any]] = field(default_factory=list)

    @property
    def is_active(self) -> bool:
        """Check if span is still active."""
        return self.end_time is None

    @property
    def duration_ms(self) -> float | None:
        """Get span duration in milliseconds."""
        if self.end_time is None:
            return None
        delta = self.end_time - self.start_time
        return delta.total_seconds() * 1000

    def set_attribute(self, key: str, value: Any) -> None:
        """Set a span attribute."""
        self.attributes[key] = value

    def add_event(
        self,
        name: str,
        attributes: dict[str, Any] | None = None,
    ) -> None:
        """Add an event to the span.

        Args:
            name: Event name.
            attributes: Event attributes.
        """
        self.events.append({
            "name": name,
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "attributes": attributes or {},
        })

    def set_status(self, status: SpanStatus, message: str = "") -> None:
        """Set span status.

        Args:
            status: New status.
            message: Optional status message.
        """
        self.status = status
        if message:
            self.attributes["status_message"] = message

    def end(self) -> None:
        """End the span."""
        if self.end_time is None:
            self.end_time = datetime.now(timezone.utc)

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            "trace_id": self.trace_id,
            "span_id": self.span_id,
            "parent_span_id": self.parent_span_id,
            "name": self.name,
            "start_time": self.start_time.isoformat(),
            "end_time": self.end_time.isoformat() if self.end_time else None,
            "duration_ms": self.duration_ms,
            "status": self.status.value,
            "attributes": self.attributes,
            "events": self.events,
        }


# =============================================================================
# Trace Context
# =============================================================================


@dataclass
class TraceContext:
    """Context for a complete trace.

    A trace represents a complete request flow and contains multiple spans.

    Attributes:
        trace_id: Unique trace identifier.
        root_span: The root span of the trace.
        baggage: Key-value pairs propagated across service boundaries.
    """

    trace_id: str
    root_span: SpanContext | None = None
    baggage: dict[str, str] = field(default_factory=dict)

    @property
    def span_id(self) -> str | None:
        """Get current span ID."""
        return self.root_span.span_id if self.root_span else None

    def set_baggage(self, key: str, value: str) -> None:
        """Set a baggage item.

        Baggage is propagated to all downstream services.

        Args:
            key: Baggage key.
            value: Baggage value.
        """
        self.baggage[key] = value

    def get_baggage(self, key: str) -> str | None:
        """Get a baggage item.

        Args:
            key: Baggage key.

        Returns:
            Baggage value or None.
        """
        return self.baggage.get(key)

    def to_headers(self) -> dict[str, str]:
        """Convert to W3C Trace Context headers.

        Returns:
            Dictionary with traceparent and tracestate headers.
        """
        headers = {}

        if self.root_span:
            # W3C traceparent format: version-trace_id-span_id-flags
            traceparent = f"00-{self.trace_id}-{self.root_span.span_id}-01"
            headers["traceparent"] = traceparent

        if self.baggage:
            # W3C baggage format: key=value,key2=value2
            baggage_str = ",".join(f"{k}={v}" for k, v in self.baggage.items())
            headers["baggage"] = baggage_str

        return headers

    @classmethod
    def from_headers(cls, headers: dict[str, str]) -> "TraceContext | None":
        """Parse trace context from W3C headers.

        Args:
            headers: HTTP headers dictionary.

        Returns:
            TraceContext or None if no valid headers.
        """
        traceparent = headers.get("traceparent")
        if not traceparent:
            return None

        try:
            parts = traceparent.split("-")
            if len(parts) != 4:
                return None

            version, trace_id, parent_span_id, flags = parts

            # Create context
            ctx = cls(trace_id=trace_id)

            # Parse baggage
            baggage_str = headers.get("baggage", "")
            if baggage_str:
                for item in baggage_str.split(","):
                    if "=" in item:
                        key, value = item.split("=", 1)
                        ctx.baggage[key.strip()] = value.strip()

            # Note: parent_span_id is stored but root_span is created by caller
            ctx._parent_span_id = parent_span_id

            return ctx

        except Exception:
            return None


# =============================================================================
# Context Storage
# =============================================================================


class _ContextStorage:
    """Thread-local storage for trace context."""

    _local = threading.local()

    @classmethod
    def get_context(cls) -> TraceContext | None:
        """Get current trace context."""
        return getattr(cls._local, "context", None)

    @classmethod
    def set_context(cls, context: TraceContext | None) -> None:
        """Set current trace context."""
        cls._local.context = context

    @classmethod
    def get_span_stack(cls) -> list[SpanContext]:
        """Get current span stack."""
        if not hasattr(cls._local, "span_stack"):
            cls._local.span_stack = []
        return cls._local.span_stack

    @classmethod
    def push_span(cls, span: SpanContext) -> None:
        """Push a span onto the stack."""
        stack = cls.get_span_stack()
        stack.append(span)

    @classmethod
    def pop_span(cls) -> SpanContext | None:
        """Pop a span from the stack."""
        stack = cls.get_span_stack()
        return stack.pop() if stack else None

    @classmethod
    def current_span(cls) -> SpanContext | None:
        """Get current span."""
        stack = cls.get_span_stack()
        return stack[-1] if stack else None


# =============================================================================
# Context Management Functions
# =============================================================================


def current_context() -> TraceContext | None:
    """Get the current trace context.

    Returns:
        Current TraceContext or None.
    """
    return _ContextStorage.get_context()


def current_span() -> SpanContext | None:
    """Get the current span.

    Returns:
        Current SpanContext or None.
    """
    return _ContextStorage.current_span()


@contextmanager
def with_context(context: TraceContext) -> Iterator[TraceContext]:
    """Context manager to set trace context.

    Args:
        context: TraceContext to use.

    Yields:
        The trace context.
    """
    previous = _ContextStorage.get_context()
    _ContextStorage.set_context(context)
    try:
        yield context
    finally:
        _ContextStorage.set_context(previous)


def create_trace(
    name: str = "",
    *,
    trace_id: str | None = None,
    attributes: dict[str, Any] | None = None,
) -> TraceContext:
    """Create a new trace with root span.

    Args:
        name: Name for the root span.
        trace_id: Custom trace ID (generates new if None).
        attributes: Initial span attributes.

    Returns:
        New TraceContext.
    """
    tid = trace_id or generate_trace_id()

    root_span = SpanContext(
        trace_id=tid,
        span_id=generate_span_id(),
        name=name,
        attributes=attributes or {},
    )

    return TraceContext(
        trace_id=tid,
        root_span=root_span,
    )


@contextmanager
def create_span(
    name: str,
    *,
    attributes: dict[str, Any] | None = None,
) -> Iterator[SpanContext]:
    """Create a new span within current trace.

    Args:
        name: Span name.
        attributes: Span attributes.

    Yields:
        The new SpanContext.
    """
    context = _ContextStorage.get_context()
    parent = _ContextStorage.current_span()

    if context is None:
        # No active trace, create one
        context = create_trace(name)
        _ContextStorage.set_context(context)
        span = context.root_span
    else:
        # Create child span
        span = SpanContext(
            trace_id=context.trace_id,
            span_id=generate_span_id(),
            parent_span_id=parent.span_id if parent else None,
            name=name,
            attributes=attributes or {},
        )

    _ContextStorage.push_span(span)

    try:
        yield span
        if span.status == SpanStatus.UNSET:
            span.set_status(SpanStatus.OK)
    except Exception as e:
        span.set_status(SpanStatus.ERROR, str(e))
        span.set_attribute("exception.type", type(e).__name__)
        span.set_attribute("exception.message", str(e))
        raise
    finally:
        span.end()
        _ContextStorage.pop_span()


# =============================================================================
# Span Collectors
# =============================================================================


class SpanCollector:
    """Collects completed spans for export.

    SpanCollector receives completed spans and can export them
    to various backends (console, file, tracing services).
    """

    def __init__(self) -> None:
        """Initialize collector."""
        self._spans: list[SpanContext] = []
        self._lock = threading.Lock()
        self._handlers: list[Callable[[SpanContext], None]] = []

    def add_handler(self, handler: Callable[[SpanContext], None]) -> None:
        """Add a span handler.

        Args:
            handler: Callable that receives completed spans.
        """
        self._handlers.append(handler)

    def collect(self, span: SpanContext) -> None:
        """Collect a completed span.

        Args:
            span: Completed span.
        """
        with self._lock:
            self._spans.append(span)

        for handler in self._handlers:
            try:
                handler(span)
            except Exception:
                pass

    def get_spans(self) -> list[SpanContext]:
        """Get all collected spans."""
        with self._lock:
            return list(self._spans)

    def clear(self) -> None:
        """Clear collected spans."""
        with self._lock:
            self._spans.clear()


# Global span collector
_span_collector = SpanCollector()


def get_span_collector() -> SpanCollector:
    """Get the global span collector."""
    return _span_collector


# =============================================================================
# Propagation Helpers
# =============================================================================


def inject_context(
    headers: dict[str, str],
    context: TraceContext | None = None,
) -> dict[str, str]:
    """Inject trace context into headers.

    Args:
        headers: Headers dictionary to modify.
        context: TraceContext (uses current if None).

    Returns:
        Headers with trace context.
    """
    ctx = context or current_context()
    if ctx:
        headers.update(ctx.to_headers())
    return headers


def extract_context(headers: dict[str, str]) -> TraceContext | None:
    """Extract trace context from headers.

    Args:
        headers: HTTP headers.

    Returns:
        TraceContext or None.
    """
    return TraceContext.from_headers(headers)



"""Ray Backend for Distributed Checkpoints.

This backend provides distributed execution using Ray,
a unified framework for scaling AI and Python applications.

Requirements:
    pip install ray
"""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any

from truthound.checkpoint.distributed.base import BaseDistributedBackend
from truthound.checkpoint.distributed.protocols import (
    BackendCapability,
    BackendNotAvailableError,
    ClusterState,
    DistributedConfig,
    DistributedTask,
    DistributedTaskProtocol,
    TaskPriority,
    TaskState,
    WorkerInfo,
    WorkerState,
)

if TYPE_CHECKING:
    from truthound.checkpoint.checkpoint import Checkpoint, CheckpointResult


logger = logging.getLogger(__name__)


# Check if Ray is available
try:
    import ray
    from ray.exceptions import RayTaskError, GetTimeoutError
    RAY_AVAILABLE = True
except ImportError:
    RAY_AVAILABLE = False
    ray = None


@dataclass
class RayBackendConfig:
    """Configuration for Ray backend.

    Attributes:
        address: Ray cluster address (None = start local cluster).
        num_cpus: Number of CPUs for local cluster.
        num_gpus: Number of GPUs for local cluster.
        runtime_env: Ray runtime environment.
        namespace: Ray namespace.
        object_store_memory: Object store memory limit.
        dashboard_host: Dashboard host.
        dashboard_port: Dashboard port.
    """

    address: str | None = None
    num_cpus: int | None = None
    num_gpus: int | None = None
    runtime_env: dict[str, Any] | None = None
    namespace: str = "truthound"
    object_store_memory: int | None = None
    dashboard_host: str = "0.0.0.0"
    dashboard_port: int = 8265


class RayTask(DistributedTask["CheckpointResult"]):
    """Ray-specific task wrapper."""

    def __init__(
        self,
        task_id: str,
        checkpoint_name: str,
        object_ref: "ray.ObjectRef",
        backend: "RayBackend",
    ) -> None:
        super().__init__(
            task_id=task_id,
            checkpoint_name=checkpoint_name,
            _backend=backend,
        )
        self._object_ref = object_ref
        self._result_cache: "CheckpointResult | None" = None
        self._error_cache: Exception | None = None

    @property
    def state(self) -> TaskState:
        """Get task state by checking Ray object status."""
        if self._result_cache is not None:
            return TaskState.SUCCEEDED
        if self._error_cache is not None:
            return TaskState.FAILED

        # Check if ready
        ready, _ = ray.wait([self._object_ref], timeout=0)
        if ready:
            try:
                self._result_cache = ray.get(self._object_ref, timeout=0)
                return TaskState.SUCCEEDED
            except Exception as e:
                self._error_cache = e
                return TaskState.FAILED

        return TaskState.RUNNING

    def result(self, timeout: float | None = None) -> "CheckpointResult":
        """Wait for and return the result."""
        if self._result_cache is not None:
            return self._result_cache

        try:
            result_dict = ray.get(self._object_ref, timeout=timeout)
            from truthound.checkpoint.checkpoint import CheckpointResult
            self._result_cache = CheckpointResult.from_dict(result_dict)
            return self._result_cache
        except GetTimeoutError:
            from truthound.checkpoint.distributed.protocols import TaskTimeoutError
            raise TaskTimeoutError(
                f"Task {self.task_id} timed out",
                self.task_id,
                timeout or 0,
            )
        except RayTaskError as e:
            self._error_cache = e
            raise

    async def result_async(self, timeout: float | None = None) -> "CheckpointResult":
        """Async version of result()."""
        import asyncio

        if self._result_cache is not None:
            return self._result_cache

        try:
            # Ray doesn't have native async get, use executor
            loop = asyncio.get_event_loop()
            result_dict = await loop.run_in_executor(
                None,
                lambda: ray.get(self._object_ref, timeout=timeout),
            )
            from truthound.checkpoint.checkpoint import CheckpointResult
            self._result_cache = CheckpointResult.from_dict(result_dict)
            return self._result_cache
        except GetTimeoutError:
            from truthound.checkpoint.distributed.protocols import TaskTimeoutError
            raise TaskTimeoutError(
                f"Task {self.task_id} timed out",
                self.task_id,
                timeout or 0,
            )

    def cancel(self, terminate: bool = False) -> bool:
        """Cancel the Ray task."""
        try:
            ray.cancel(self._object_ref, force=terminate)
            return True
        except Exception as e:
            logger.error(f"Failed to cancel task: {e}")
            return False

    def is_ready(self) -> bool:
        """Check if task is complete."""
        if self._result_cache or self._error_cache:
            return True
        ready, _ = ray.wait([self._object_ref], timeout=0)
        return len(ready) > 0

    def is_successful(self) -> bool:
        """Check if task succeeded."""
        if self._error_cache:
            return False
        if self._result_cache:
            return True
        if self.is_ready():
            try:
                self.result(timeout=0)
                return True
            except Exception:
                return False
        return False


class RayBackend(BaseDistributedBackend):
    """Ray-based distributed backend.

    This backend uses Ray for distributed task execution, providing
    automatic scaling and fault tolerance.

    Example:
        >>> from truthound.checkpoint.distributed import RayBackend
        >>>
        >>> # Connect to existing cluster
        >>> backend = RayBackend(address="ray://localhost:10001")
        >>>
        >>> # Or start local cluster
        >>> backend = RayBackend(num_cpus=4)
        >>>
        >>> with backend.connection():
        ...     task = backend.submit(checkpoint)
        ...     result = task.result(timeout=300)
    """

    def __init__(
        self,
        config: DistributedConfig | None = None,
        address: str | None = None,
        num_cpus: int | None = None,
        num_gpus: int | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize Ray backend.

        Args:
            config: Distributed configuration.
            address: Ray cluster address.
            num_cpus: Number of CPUs for local cluster.
            num_gpus: Number of GPUs for local cluster.
            **kwargs: Additional Ray configuration.
        """
        if not RAY_AVAILABLE:
            raise BackendNotAvailableError(
                "ray",
                reason="Ray is not installed",
                install_hint="pip install ray",
            )

        super().__init__(config, **kwargs)

        self._ray_config = RayBackendConfig(
            address=address,
            num_cpus=num_cpus,
            num_gpus=num_gpus,
        )

        # Apply additional kwargs
        for key, value in kwargs.items():
            if hasattr(self._ray_config, key):
                setattr(self._ray_config, key, value)

        self._initialized_ray = False

    @property
    def name(self) -> str:
        return "ray"

    @property
    def capabilities(self) -> BackendCapability:
        return (
            BackendCapability.ASYNC_SUBMIT
            | BackendCapability.BATCH_SUBMIT
            | BackendCapability.RESULT_BACKEND
            | BackendCapability.TASK_REVOKE
            | BackendCapability.WORKER_SCALING
            | BackendCapability.HEALTH_CHECK
            | BackendCapability.METRICS
        )

    def _do_connect(self, **kwargs: Any) -> None:
        """Initialize Ray runtime."""
        if ray.is_initialized():
            logger.info("Ray already initialized, reusing existing context")
            return

        init_kwargs = {
            "namespace": self._ray_config.namespace,
        }

        if self._ray_config.address:
            init_kwargs["address"] = self._ray_config.address
        else:
            # Local cluster
            if self._ray_config.num_cpus:
                init_kwargs["num_cpus"] = self._ray_config.num_cpus
            if self._ray_config.num_gpus:
                init_kwargs["num_gpus"] = self._ray_config.num_gpus
            if self._ray_config.object_store_memory:
                init_kwargs["object_store_memory"] = self._ray_config.object_store_memory

        if self._ray_config.runtime_env:
            init_kwargs["runtime_env"] = self._ray_config.runtime_env

        # Apply override kwargs
        init_kwargs.update(kwargs)

        ray.init(**init_kwargs)
        self._initialized_ray = True

        logger.info(f"Ray backend connected: {ray.get_runtime_context().get_node_id()}")

    def _do_disconnect(self) -> None:
        """Shutdown Ray if we initialized it."""
        if self._initialized_ray and ray.is_initialized():
            ray.shutdown()
            self._initialized_ray = False
            logger.info("Ray backend disconnected")

    def _do_submit(
        self,
        task: DistributedTask["CheckpointResult"],
        checkpoint: "Checkpoint",
        priority: TaskPriority,
        timeout: float,
        context: dict[str, Any] | None,
        **kwargs: Any,
    ) -> None:
        """Submit task to Ray."""
        # Create remote function
        @ray.remote
        def execute_checkpoint(
            checkpoint_dict: dict[str, Any],
            context: dict[str, Any] | None,
        ) -> dict[str, Any]:
            """Execute checkpoint in Ray worker."""
            from truthound.checkpoint.checkpoint import (
                Checkpoint,
                CheckpointConfig,
            )

            config = CheckpointConfig(**checkpoint_dict.get("config", {}))
            cp = Checkpoint(config=config)
            result = cp.run(context=context)

            return result.to_dict()

        # Serialize checkpoint
        checkpoint_dict = {
            "config": {
                "name": checkpoint.config.name,
                "data_source": str(checkpoint.config.data_source),
                "validators": checkpoint.config.validators,
                "min_severity": checkpoint.config.min_severity,
                "fail_on_critical": checkpoint.config.fail_on_critical,
                "fail_on_high": checkpoint.config.fail_on_high,
                "timeout_seconds": checkpoint.config.timeout_seconds,
                "sample_size": checkpoint.config.sample_size,
            }
        }

        # Determine resources
        num_cpus = kwargs.get("num_cpus", 1)
        num_gpus = kwargs.get("num_gpus", 0)

        # Submit to Ray
        object_ref = execute_checkpoint.options(
            num_cpus=num_cpus,
            num_gpus=num_gpus,
        ).remote(checkpoint_dict, context)

        # Create Ray task wrapper
        ray_task = RayTask(
            task_id=f"ray-{object_ref.hex()[:16]}",
            checkpoint_name=checkpoint.name,
            object_ref=object_ref,
            backend=self,
        )

        # Update original task
        task.task_id = ray_task.task_id
        task._object_ref = object_ref  # type: ignore
        task._set_state(TaskState.RUNNING)
        task._started_at = datetime.now()

        # Update tracking
        with self._lock:
            self._tasks[task.task_id] = task

    def submit_batch(
        self,
        checkpoints: list["Checkpoint"],
        priority: TaskPriority = TaskPriority.NORMAL,
        timeout: float | None = None,
        context: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[DistributedTaskProtocol["CheckpointResult"]]:
        """Submit multiple checkpoints efficiently using Ray's parallelism."""
        tasks = []

        # Create remote function once
        @ray.remote
        def execute_checkpoint_batch(
            checkpoint_dict: dict[str, Any],
            ctx: dict[str, Any] | None,
        ) -> dict[str, Any]:
            from truthound.checkpoint.checkpoint import (
                Checkpoint,
                CheckpointConfig,
            )

            config = CheckpointConfig(**checkpoint_dict.get("config", {}))
            cp = Checkpoint(config=config)
            result = cp.run(context=ctx)

            return result.to_dict()

        # Submit all at once
        object_refs = []
        for checkpoint in checkpoints:
            checkpoint_dict = {
                "config": {
                    "name": checkpoint.config.name,
                    "data_source": str(checkpoint.config.data_source),
                    "validators": checkpoint.config.validators,
                    "min_severity": checkpoint.config.min_severity,
                    "fail_on_critical": checkpoint.config.fail_on_critical,
                    "fail_on_high": checkpoint.config.fail_on_high,
                    "timeout_seconds": checkpoint.config.timeout_seconds,
                    "sample_size": checkpoint.config.sample_size,
                }
            }

            ref = execute_checkpoint_batch.remote(checkpoint_dict, context)
            object_refs.append((checkpoint, ref))

        # Create task wrappers
        for checkpoint, ref in object_refs:
            task: DistributedTask["CheckpointResult"] = DistributedTask.create(
                checkpoint=checkpoint,
                backend=self,
            )

            ray_task = RayTask(
                task_id=f"ray-{ref.hex()[:16]}",
                checkpoint_name=checkpoint.name,
                object_ref=ref,
                backend=self,
            )

            task.task_id = ray_task.task_id
            task._object_ref = ref  # type: ignore
            task._set_state(TaskState.RUNNING)
            task._started_at = datetime.now()

            with self._lock:
                self._tasks[task.task_id] = task

            tasks.append(task)
            self._metrics.record_submission()

        return tasks

    def _do_get_cluster_state(self) -> ClusterState:
        """Get Ray cluster state."""
        if not ray.is_initialized():
            return ClusterState(backend_name=self.name, is_healthy=False)

        try:
            nodes = ray.nodes()
            workers = []

            for node in nodes:
                if not node.get("Alive", False):
                    continue

                resources = node.get("Resources", {})
                resources_used = node.get("Resources_Used", {})

                total_cpus = resources.get("CPU", 0)
                used_cpus = resources_used.get("CPU", 0)

                workers.append(WorkerInfo(
                    worker_id=node.get("NodeID", "unknown"),
                    hostname=node.get("NodeManagerAddress", "unknown"),
                    state=WorkerState.ONLINE if node.get("Alive") else WorkerState.OFFLINE,
                    current_tasks=int(used_cpus),
                    max_concurrency=int(total_cpus),
                    metadata={
                        "object_store_memory": node.get("ObjectStoreMemory"),
                        "ray_version": node.get("RayVersion"),
                    },
                ))

            total_capacity = sum(w.max_concurrency for w in workers)
            current_load = sum(w.current_tasks for w in workers)

            return ClusterState(
                workers=workers,
                total_capacity=total_capacity,
                current_load=current_load,
                pending_tasks=0,  # Ray doesn't expose pending task count easily
                backend_name=self.name,
                backend_version=ray.__version__,
                is_healthy=len(workers) > 0,
            )

        except Exception as e:
            logger.error(f"Failed to get cluster state: {e}")
            return ClusterState(backend_name=self.name, is_healthy=False)

    def _do_get_workers(self) -> list[WorkerInfo]:
        """Get list of Ray nodes."""
        state = self._do_get_cluster_state()
        return state.workers

    def _do_scale_workers(self, count: int) -> bool:
        """Scale Ray cluster (requires Ray autoscaler)."""
        # Ray autoscaling is typically managed externally
        # This would require Ray cluster configuration
        logger.warning("Ray autoscaling requires external cluster management")
        return False


class RayActorBackend(RayBackend):
    """Ray backend using persistent actors for better performance.

    This variant uses Ray actors to maintain checkpoint execution context,
    reducing initialization overhead for repeated executions.
    """

    def __init__(
        self,
        config: DistributedConfig | None = None,
        num_actors: int = 4,
        **kwargs: Any,
    ) -> None:
        super().__init__(config, **kwargs)
        self._num_actors = num_actors
        self._actors: list[Any] = []
        self._actor_index = 0
        self._actor_lock = None

    def _do_connect(self, **kwargs: Any) -> None:
        """Initialize Ray and create actor pool."""
        import threading
        super()._do_connect(**kwargs)

        self._actor_lock = threading.Lock()

        # Create actor class
        @ray.remote
        class CheckpointExecutor:
            def __init__(self) -> None:
                self._execution_count = 0

            def execute(
                self,
                checkpoint_dict: dict[str, Any],
                context: dict[str, Any] | None,
            ) -> dict[str, Any]:
                from truthound.checkpoint.checkpoint import (
                    Checkpoint,
                    CheckpointConfig,
                )

                config = CheckpointConfig(**checkpoint_dict.get("config", {}))
                checkpoint = Checkpoint(config=config)
                result = checkpoint.run(context=context)

                self._execution_count += 1

                return result.to_dict()

            def get_stats(self) -> dict[str, Any]:
                return {"execution_count": self._execution_count}

        # Create actor pool
        self._actors = [CheckpointExecutor.remote() for _ in range(self._num_actors)]

        logger.info(f"Created {self._num_actors} Ray checkpoint executor actors")

    def _do_disconnect(self) -> None:
        """Shutdown actors and Ray."""
        for actor in self._actors:
            try:
                ray.kill(actor)
            except Exception:
                pass
        self._actors.clear()
        super()._do_disconnect()

    def _get_next_actor(self) -> Any:
        """Get next actor in round-robin fashion."""
        with self._actor_lock:
            actor = self._actors[self._actor_index]
            self._actor_index = (self._actor_index + 1) % len(self._actors)
            return actor

    def _do_submit(
        self,
        task: DistributedTask["CheckpointResult"],
        checkpoint: "Checkpoint",
        priority: TaskPriority,
        timeout: float,
        context: dict[str, Any] | None,
        **kwargs: Any,
    ) -> None:
        """Submit task to an actor."""
        actor = self._get_next_actor()

        checkpoint_dict = {
            "config": {
                "name": checkpoint.config.name,
                "data_source": str(checkpoint.config.data_source),
                "validators": checkpoint.config.validators,
                "min_severity": checkpoint.config.min_severity,
                "fail_on_critical": checkpoint.config.fail_on_critical,
                "fail_on_high": checkpoint.config.fail_on_high,
                "timeout_seconds": checkpoint.config.timeout_seconds,
                "sample_size": checkpoint.config.sample_size,
            }
        }

        # Submit to actor
        object_ref = actor.execute.remote(checkpoint_dict, context)

        ray_task = RayTask(
            task_id=f"ray-actor-{object_ref.hex()[:16]}",
            checkpoint_name=checkpoint.name,
            object_ref=object_ref,
            backend=self,
        )

        task.task_id = ray_task.task_id
        task._object_ref = object_ref  # type: ignore
        task._set_state(TaskState.RUNNING)
        task._started_at = datetime.now()

        with self._lock:
            self._tasks[task.task_id] = task



"""Base classes for table metadata validators.

This module provides base classes for table-level validation.
"""

from abc import abstractmethod
from typing import Any

import polars as pl

from truthound.types import Severity
from truthound.validators.base import ValidationIssue, Validator


class TableValidator(Validator):
    """Base class for table-level validators.

    Table validators check metadata and aggregate properties of entire tables,
    such as row counts, column counts, freshness, and schema matching.
    """

    name = "table_base"
    category = "table"

    def __init__(self, **kwargs: Any):
        """Initialize table validator.

        Args:
            **kwargs: Additional config passed to base Validator
        """
        super().__init__(**kwargs)

    @abstractmethod
    def validate(self, lf: pl.LazyFrame) -> list[ValidationIssue]:
        """Validate the table.

        Args:
            lf: LazyFrame to validate

        Returns:
            List of validation issues
        """
        pass


class TableSchemaValidator(Validator):
    """Base class for schema-focused table validators.

    Provides common functionality for validators that check table structure.
    """

    name = "table_schema_base"
    category = "table"

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def _get_schema(self, lf: pl.LazyFrame) -> dict[str, pl.DataType]:
        """Get schema of the LazyFrame.

        Args:
            lf: LazyFrame to get schema from

        Returns:
            Dictionary of column name to data type
        """
        return lf.collect_schema()

    def _get_columns(self, lf: pl.LazyFrame) -> list[str]:
        """Get column names of the LazyFrame.

        Args:
            lf: LazyFrame to get columns from

        Returns:
            List of column names
        """
        return list(lf.collect_schema().keys())



"""Process-isolated timeout execution for reliable operation termination.

This module provides a robust timeout system that can reliably terminate
operations including native Rust code (like Polars) by using process isolation.

Key features:
- Process-based execution for reliable termination
- Pre-execution complexity estimation
- Circuit breaker pattern for repeated failures
- Resource monitoring (memory, CPU)
- Graceful degradation with multiple strategies

Problem Solved:
- Python threading cannot interrupt Polars Rust operations
- Process isolation ensures SIGTERM/SIGKILL always works

Design Principles:
- Strategy Pattern: Multiple execution backends (thread, process, async)
- Circuit Breaker: Prevent cascade failures
- Bulkhead: Isolate resources per operation
- Fail-Fast: Pre-check before expensive operations

Example:
    from truthound.profiler.process_timeout import (
        ProcessTimeoutExecutor,
        TimeoutConfig,
        with_process_timeout,
    )

    # Basic usage
    executor = ProcessTimeoutExecutor()
    result = executor.execute(
        expensive_polars_operation,
        timeout_seconds=30,
    )

    # With complexity estimation
    result = executor.execute_with_estimation(
        operation,
        data_size=1_000_000,
        timeout_seconds=60,
    )
"""

from __future__ import annotations

import functools
import hashlib
import logging
import multiprocessing as mp
import os
import pickle
import queue
import signal
import sys
import threading
import time
import traceback
from abc import ABC, abstractmethod
from concurrent.futures import (
    Future,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    TimeoutError as FuturesTimeoutError,
)
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from multiprocessing import Queue
from typing import (
    Any,
    Callable,
    Generic,
    Generator,
    Protocol,
    TypeVar,
)

logger = logging.getLogger(__name__)


# =============================================================================
# Types
# =============================================================================

T = TypeVar("T")
R = TypeVar("R")


class ExecutionBackend(str, Enum):
    """Available execution backends."""

    THREAD = "thread"       # Fast but can't interrupt native code
    PROCESS = "process"     # Reliable but has serialization overhead
    ADAPTIVE = "adaptive"   # Auto-select based on operation type
    INLINE = "inline"       # No isolation (for debugging)


class TimeoutAction(str, Enum):
    """Actions to take when timeout occurs."""

    SKIP = "skip"           # Skip and continue
    PARTIAL = "partial"     # Return partial results
    FAIL = "fail"           # Raise exception
    RETRY = "retry"         # Retry with extended timeout
    CIRCUIT_BREAK = "circuit_break"  # Open circuit breaker


class TerminationMethod(str, Enum):
    """Methods for terminating processes."""

    GRACEFUL = "graceful"   # SIGTERM, wait, then SIGKILL
    IMMEDIATE = "immediate"  # SIGKILL directly
    COOPERATIVE = "cooperative"  # Set flag and wait


class CircuitState(str, Enum):
    """Circuit breaker states."""

    CLOSED = "closed"       # Normal operation
    OPEN = "open"           # Rejecting requests
    HALF_OPEN = "half_open"  # Testing if recovered


# =============================================================================
# Execution Result
# =============================================================================


@dataclass
class ExecutionMetrics:
    """Metrics from an execution attempt."""

    started_at: datetime = field(default_factory=datetime.now)
    completed_at: datetime | None = None
    elapsed_seconds: float = 0.0
    peak_memory_mb: float = 0.0
    cpu_time_seconds: float = 0.0
    backend_used: ExecutionBackend = ExecutionBackend.THREAD
    was_terminated: bool = False
    termination_method: TerminationMethod | None = None
    retries: int = 0

    def complete(self) -> None:
        """Mark execution as complete."""
        self.completed_at = datetime.now()
        self.elapsed_seconds = (self.completed_at - self.started_at).total_seconds()


@dataclass
class ExecutionResult(Generic[T]):
    """Result of a timed execution.

    Attributes:
        success: Whether operation completed successfully
        value: Result value if successful
        error: Exception if failed
        timed_out: Whether operation was terminated due to timeout
        metrics: Execution metrics
        partial_result: Partial result if available
    """

    success: bool
    value: T | None = None
    error: Exception | None = None
    timed_out: bool = False
    metrics: ExecutionMetrics = field(default_factory=ExecutionMetrics)
    partial_result: Any = None

    @classmethod
    def ok(cls, value: T, metrics: ExecutionMetrics | None = None) -> "ExecutionResult[T]":
        """Create successful result."""
        m = metrics or ExecutionMetrics()
        m.complete()
        return cls(success=True, value=value, metrics=m)

    @classmethod
    def timeout(
        cls,
        metrics: ExecutionMetrics | None = None,
        partial: Any = None,
    ) -> "ExecutionResult[T]":
        """Create timeout result."""
        m = metrics or ExecutionMetrics()
        m.complete()
        m.was_terminated = True
        return cls(success=False, timed_out=True, metrics=m, partial_result=partial)

    @classmethod
    def failure(
        cls,
        error: Exception,
        metrics: ExecutionMetrics | None = None,
    ) -> "ExecutionResult[T]":
        """Create failure result."""
        m = metrics or ExecutionMetrics()
        m.complete()
        return cls(success=False, error=error, metrics=m)

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        return {
            "success": self.success,
            "timed_out": self.timed_out,
            "elapsed_seconds": self.metrics.elapsed_seconds,
            "backend_used": self.metrics.backend_used.value,
            "was_terminated": self.metrics.was_terminated,
            "error": str(self.error) if self.error else None,
        }


# =============================================================================
# Complexity Estimator
# =============================================================================


@dataclass
class ComplexityEstimate:
    """Estimated complexity of an operation.

    Attributes:
        estimated_time_seconds: Expected execution time
        estimated_memory_mb: Expected memory usage
        confidence: Confidence in the estimate (0-1)
        recommendation: Recommended execution backend
        should_sample: Whether to sample data first
        max_safe_rows: Maximum rows to process safely
    """

    estimated_time_seconds: float
    estimated_memory_mb: float
    confidence: float = 0.5
    recommendation: ExecutionBackend = ExecutionBackend.ADAPTIVE
    should_sample: bool = False
    max_safe_rows: int = 0
    risk_level: str = "unknown"

    def exceeds_timeout(self, timeout_seconds: float) -> bool:
        """Check if estimated time exceeds timeout."""
        # Use confidence-adjusted estimate
        adjusted = self.estimated_time_seconds * (2 - self.confidence)
        return adjusted > timeout_seconds

    def exceeds_memory(self, max_memory_mb: float) -> bool:
        """Check if estimated memory exceeds limit."""
        adjusted = self.estimated_memory_mb * (2 - self.confidence)
        return adjusted > max_memory_mb


class ComplexityEstimator(Protocol):
    """Protocol for complexity estimation."""

    def estimate(
        self,
        operation_type: str,
        data_size: int,
        column_count: int = 1,
        **kwargs: Any,
    ) -> ComplexityEstimate:
        """Estimate operation complexity."""
        ...


class DefaultComplexityEstimator:
    """Default complexity estimator using heuristics.

    Uses empirical constants calibrated for common operations.
    """

    # Calibration constants (rows per second for different operations)
    OPERATION_SPEEDS: dict[str, float] = {
        "profile_column": 100_000,      # 100K rows/sec
        "pattern_match": 50_000,        # 50K rows/sec
        "distribution": 200_000,        # 200K rows/sec
        "correlation": 10_000,          # 10K rows/sec (O(n²))
        "unique_count": 150_000,        # 150K rows/sec
        "null_check": 500_000,          # 500K rows/sec
        "type_inference": 75_000,       # 75K rows/sec
        "default": 100_000,             # Default fallback
    }

    # Memory constants (bytes per row for different operations)
    MEMORY_PER_ROW: dict[str, float] = {
        "profile_column": 100,          # 100 bytes/row
        "pattern_match": 200,           # 200 bytes/row (regex buffers)
        "distribution": 50,             # 50 bytes/row
        "correlation": 300,             # 300 bytes/row (matrix)
        "unique_count": 150,            # 150 bytes/row (hash set)
        "default": 100,
    }

    def __init__(self, safety_factor: float = 1.5):
        """Initialize estimator.

        Args:
            safety_factor: Multiplier for conservative estimates
        """
        self.safety_factor = safety_factor
        self._history: list[tuple[str, int, float]] = []  # For calibration

    def estimate(
        self,
        operation_type: str,
        data_size: int,
        column_count: int = 1,
        **kwargs: Any,
    ) -> ComplexityEstimate:
        """Estimate operation complexity.

        Args:
            operation_type: Type of operation
            data_size: Number of rows
            column_count: Number of columns
            **kwargs: Additional hints

        Returns:
            Complexity estimate
        """
        # Get calibration constants
        speed = self.OPERATION_SPEEDS.get(
            operation_type,
            self.OPERATION_SPEEDS["default"],
        )
        memory_per_row = self.MEMORY_PER_ROW.get(
            operation_type,
            self.MEMORY_PER_ROW["default"],
        )

        # Calculate estimates
        time_estimate = (data_size / speed) * column_count * self.safety_factor
        memory_estimate = (data_size * memory_per_row * column_count) / (1024 * 1024)  # MB

        # Determine confidence based on data size
        if data_size < 10_000:
            confidence = 0.9  # High confidence for small data
        elif data_size < 100_000:
            confidence = 0.7
        elif data_size < 1_000_000:
            confidence = 0.5
        else:
            confidence = 0.3  # Low confidence for very large data

        # Determine recommendation
        if time_estimate > 60 or memory_estimate > 1000:
            recommendation = ExecutionBackend.PROCESS
            should_sample = True
            risk_level = "high"
        elif time_estimate > 10 or memory_estimate > 500:
            recommendation = ExecutionBackend.PROCESS
            should_sample = False
            risk_level = "medium"
        else:
            recommendation = ExecutionBackend.THREAD
            should_sample = False
            risk_level = "low"

        # Calculate safe row limit
        max_safe_rows = int(speed * 30)  # 30 seconds worth

        return ComplexityEstimate(
            estimated_time_seconds=time_estimate,
            estimated_memory_mb=memory_estimate,
            confidence=confidence,
            recommendation=recommendation,
            should_sample=should_sample,
            max_safe_rows=max_safe_rows,
            risk_level=risk_level,
        )

    def record_actual(
        self,
        operation_type: str,
        data_size: int,
        actual_seconds: float,
    ) -> None:
        """Record actual execution time for future calibration.

        Args:
            operation_type: Type of operation
            data_size: Number of rows
            actual_seconds: Actual execution time
        """
        self._history.append((operation_type, data_size, actual_seconds))

        # Keep last 100 records
        if len(self._history) > 100:
            self._history = self._history[-100:]


# Global estimator instance
default_complexity_estimator = DefaultComplexityEstimator()


# =============================================================================
# Circuit Breaker
# =============================================================================


@dataclass
class CircuitBreakerConfig:
    """Configuration for circuit breaker.

    Attributes:
        failure_threshold: Failures before opening circuit
        success_threshold: Successes before closing from half-open
        timeout_seconds: Time before trying half-open from open
        half_open_max_calls: Max calls in half-open state
    """

    failure_threshold: int = 5
    success_threshold: int = 2
    timeout_seconds: float = 60.0
    half_open_max_calls: int = 3


class CircuitBreaker:
    """Circuit breaker for preventing cascade failures.

    Implements the circuit breaker pattern:
    - CLOSED: Normal operation, track failures
    - OPEN: Reject all requests, wait for timeout
    - HALF_OPEN: Allow limited requests to test recovery

    Example:
        breaker = CircuitBreaker()

        if breaker.can_execute():
            try:
                result = operation()
                breaker.record_success()
            except Exception:
                breaker.record_failure()
                raise
        else:
            raise CircuitOpenError("Circuit is open")
    """

    def __init__(
        self,
        name: str = "default",
        config: CircuitBreakerConfig | None = None,
    ):
        """Initialize circuit breaker.

        Args:
            name: Identifier for this breaker
            config: Configuration
        """
        self.name = name
        self.config = config or CircuitBreakerConfig()
        self._state = CircuitState.CLOSED
        self._failure_count = 0
        self._success_count = 0
        self._last_failure_time: float | None = None
        self._half_open_calls = 0
        self._lock = threading.RLock()

    @property
    def state(self) -> CircuitState:
        """Get current state."""
        with self._lock:
            self._check_state_transition()
            return self._state

    @property
    def is_closed(self) -> bool:
        """Check if circuit is closed (normal)."""
        return self.state == CircuitState.CLOSED

    @property
    def is_open(self) -> bool:
        """Check if circuit is open (rejecting)."""
        return self.state == CircuitState.OPEN

    def can_execute(self) -> bool:
        """Check if execution is allowed.

        Returns:
            True if execution should proceed
        """
        with self._lock:
            self._check_state_transition()

            if self._state == CircuitState.CLOSED:
                return True

            if self._state == CircuitState.HALF_OPEN:
                if self._half_open_calls < self.config.half_open_max_calls:
                    self._half_open_calls += 1
                    return True
                return False

            # OPEN state
            return False

    def record_success(self) -> None:
        """Record a successful execution."""
        with self._lock:
            if self._state == CircuitState.HALF_OPEN:
                self._success_count += 1
                if self._success_count >= self.config.success_threshold:
                    self._close()
            elif self._state == CircuitState.CLOSED:
                # Reset failure count on success
                self._failure_count = 0

    def record_failure(self) -> None:
        """Record a failed execution."""
        with self._lock:
            self._last_failure_time = time.time()

            if self._state == CircuitState.HALF_OPEN:
                self._open()
            elif self._state == CircuitState.CLOSED:
                self._failure_count += 1
                if self._failure_count >= self.config.failure_threshold:
                    self._open()

    def reset(self) -> None:
        """Reset circuit to closed state."""
        with self._lock:
            self._close()

    def _check_state_transition(self) -> None:
        """Check and perform state transitions."""
        if self._state == CircuitState.OPEN:
            if self._last_failure_time is not None:
                elapsed = time.time() - self._last_failure_time
                if elapsed >= self.config.timeout_seconds:
                    self._half_open()

    def _open(self) -> None:
        """Transition to open state."""
        self._state = CircuitState.OPEN
        self._failure_count = 0
        self._success_count = 0
        logger.warning(f"Circuit breaker '{self.name}' opened")

    def _close(self) -> None:
        """Transition to closed state."""
        self._state = CircuitState.CLOSED
        self._failure_count = 0
        self._success_count = 0
        self._half_open_calls = 0
        logger.info(f"Circuit breaker '{self.name}' closed")

    def _half_open(self) -> None:
        """Transition to half-open state."""
        self._state = CircuitState.HALF_OPEN
        self._success_count = 0
        self._half_open_calls = 0
        logger.info(f"Circuit breaker '{self.name}' half-open")

    def get_stats(self) -> dict[str, Any]:
        """Get circuit breaker statistics."""
        with self._lock:
            return {
                "name": self.name,
                "state": self._state.value,
                "failure_count": self._failure_count,
                "success_count": self._success_count,
                "last_failure": self._last_failure_time,
            }


class CircuitBreakerRegistry:
    """Registry for circuit breakers by operation type."""

    def __init__(self) -> None:
        self._breakers: dict[str, CircuitBreaker] = {}
        self._lock = threading.RLock()

    def get(
        self,
        name: str,
        config: CircuitBreakerConfig | None = None,
    ) -> CircuitBreaker:
        """Get or create a circuit breaker."""
        with self._lock:
            if name not in self._breakers:
                self._breakers[name] = CircuitBreaker(name, config)
            return self._breakers[name]

    def reset_all(self) -> None:
        """Reset all circuit breakers."""
        with self._lock:
            for breaker in self._breakers.values():
                breaker.reset()


# Global registry
circuit_breaker_registry = CircuitBreakerRegistry()


# =============================================================================
# Execution Strategy Protocol
# =============================================================================


class ExecutionStrategy(ABC):
    """Abstract base class for execution strategies.

    Defines how an operation is executed with timeout control.
    """

    name: ExecutionBackend

    @abstractmethod
    def execute(
        self,
        func: Callable[[], T],
        timeout_seconds: float,
        **kwargs: Any,
    ) -> ExecutionResult[T]:
        """Execute function with timeout.

        Args:
            func: Function to execute
            timeout_seconds: Timeout in seconds
            **kwargs: Additional options

        Returns:
            Execution result
        """
        pass

    @abstractmethod
    def is_available(self) -> bool:
        """Check if this strategy is available."""
        pass


# =============================================================================
# Thread-Based Strategy
# =============================================================================


class ThreadExecutionStrategy(ExecutionStrategy):
    """Thread-based execution strategy.

    Fast with low overhead but cannot interrupt native code.
    Best for pure Python operations.
    """

    name = ExecutionBackend.THREAD

    def __init__(self, max_workers: int = 1):
        self.max_workers = max_workers

    def execute(
        self,
        func: Callable[[], T],
        timeout_seconds: float,
        **kwargs: Any,
    ) -> ExecutionResult[T]:
        """Execute in thread with timeout."""
        metrics = ExecutionMetrics(backend_used=self.name)

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            future: Future[T] = executor.submit(func)

            try:
                value = future.result(timeout=timeout_seconds)
                return ExecutionResult.ok(value, metrics)

            except FuturesTimeoutError:
                metrics.was_terminated = True
                # Note: Cannot actually terminate the thread
                return ExecutionResult.timeout(metrics)

            except Exception as e:
                return ExecutionResult.failure(e, metrics)

    def is_available(self) -> bool:
        return True


# =============================================================================
# Process-Based Strategy
# =============================================================================


def _process_worker(
    func_pickle: bytes,
    result_queue: Queue,
    ready_event: mp.Event,
) -> None:
    """Worker function for process execution.

    This runs in a separate process and can be terminated.

    Args:
        func_pickle: Pickled function to execute
        result_queue: Queue to put result
        ready_event: Event to signal ready
    """
    try:
        # Signal ready
        ready_event.set()

        # Deserialize and execute
        func = pickle.loads(func_pickle)
        result = func()

        # Put success result
        result_queue.put(("success", result, None))

    except Exception as e:
        # Put error result
        tb = traceback.format_exc()
        result_queue.put(("error", None, (type(e).__name__, str(e), tb)))


class ProcessExecutionStrategy(ExecutionStrategy):
    """Process-based execution strategy.

    Uses separate process for reliable termination.
    Has serialization overhead but can terminate any code.
    """

    name = ExecutionBackend.PROCESS

    def __init__(
        self,
        graceful_timeout: float = 2.0,
        start_method: str | None = None,
    ):
        """Initialize process strategy.

        Args:
            graceful_timeout: Time to wait after SIGTERM before SIGKILL
            start_method: Process start method (spawn, fork, forkserver)
        """
        self.graceful_timeout = graceful_timeout
        self.start_method = start_method or self._get_default_start_method()

    def _get_default_start_method(self) -> str:
        """Get default start method for platform."""
        if sys.platform == "darwin":
            return "spawn"  # fork is problematic on macOS
        elif sys.platform == "win32":
            return "spawn"
        else:
            return "fork"  # Faster on Linux

    def execute(
        self,
        func: Callable[[], T],
        timeout_seconds: float,
        **kwargs: Any,
    ) -> ExecutionResult[T]:
        """Execute in separate process with timeout."""
        metrics = ExecutionMetrics(backend_used=self.name)

        # Serialize function
        try:
            func_pickle = pickle.dumps(func)
        except Exception as e:
            return ExecutionResult.failure(
                ValueError(f"Cannot serialize function: {e}"),
                metrics,
            )

        # Create communication primitives
        ctx = mp.get_context(self.start_method)
        result_queue: Queue = ctx.Queue()
        ready_event = ctx.Event()

        # Start process
        process = ctx.Process(
            target=_process_worker,
            args=(func_pickle, result_queue, ready_event),
        )
        process.start()

        try:
            # Wait for process to be ready
            if not ready_event.wait(timeout=5.0):
                self._terminate_process(process, metrics)
                return ExecutionResult.timeout(metrics)

            # Wait for result with timeout
            try:
                status, value, error_info = result_queue.get(timeout=timeout_seconds)

                if status == "success":
                    return ExecutionResult.ok(value, metrics)
                else:
                    error_type, error_msg, tb = error_info
                    error = RuntimeError(f"{error_type}: {error_msg}")
                    return ExecutionResult.failure(error, metrics)

            except queue.Empty:
                # Timeout
                self._terminate_process(process, metrics)
                return ExecutionResult.timeout(metrics)

        finally:
            # Ensure process is cleaned up
            if process.is_alive():
                self._terminate_process(process, metrics)
            process.join(timeout=1.0)

    def _terminate_process(
        self,
        process: mp.Process,
        metrics: ExecutionMetrics,
    ) -> None:
        """Terminate a process gracefully then forcefully."""
        metrics.was_terminated = True

        if not process.is_alive():
            return

        # Try graceful termination
        process.terminate()
        metrics.termination_method = TerminationMethod.GRACEFUL

        # Wait for graceful exit
        process.join(timeout=self.graceful_timeout)

        # Force kill if still alive
        if process.is_alive():
            try:
                os.kill(process.pid, signal.SIGKILL)
                metrics.termination_method = TerminationMethod.IMMEDIATE
            except (ProcessLookupError, OSError):
                pass

            process.join(timeout=1.0)

    def is_available(self) -> bool:
        """Check if multiprocessing is available."""
        try:
            ctx = mp.get_context(self.start_method)
            return True
        except Exception:
            return False


# =============================================================================
# Adaptive Strategy
# =============================================================================


class AdaptiveExecutionStrategy(ExecutionStrategy):
    """Adaptive strategy that selects backend based on operation.

    Uses complexity estimation to choose between thread and process.
    """

    name = ExecutionBackend.ADAPTIVE

    def __init__(
        self,
        estimator: ComplexityEstimator | None = None,
        thread_threshold_seconds: float = 5.0,
    ):
        """Initialize adaptive strategy.

        Args:
            estimator: Complexity estimator
            thread_threshold_seconds: Use thread for operations under this
        """
        self.estimator = estimator or default_complexity_estimator
        self.thread_threshold = thread_threshold_seconds
        self._thread_strategy = ThreadExecutionStrategy()
        self._process_strategy = ProcessExecutionStrategy()

    def execute(
        self,
        func: Callable[[], T],
        timeout_seconds: float,
        operation_type: str = "default",
        data_size: int = 0,
        **kwargs: Any,
    ) -> ExecutionResult[T]:
        """Execute with adaptive backend selection."""
        # Estimate complexity
        estimate = self.estimator.estimate(
            operation_type=operation_type,
            data_size=data_size,
        )

        # Select strategy
        if estimate.recommendation == ExecutionBackend.PROCESS:
            strategy = self._process_strategy
        elif estimate.estimated_time_seconds > self.thread_threshold:
            strategy = self._process_strategy
        else:
            strategy = self._thread_strategy

        logger.debug(
            f"Adaptive strategy selected {strategy.name.value} for "
            f"{operation_type} ({data_size} rows, "
            f"est. {estimate.estimated_time_seconds:.2f}s)"
        )

        # Execute
        result = strategy.execute(func, timeout_seconds, **kwargs)

        # Record actual time for calibration
        if isinstance(self.estimator, DefaultComplexityEstimator):
            self.estimator.record_actual(
                operation_type,
                data_size,
                result.metrics.elapsed_seconds,
            )

        return result

    def is_available(self) -> bool:
        return self._thread_strategy.is_available()


# =============================================================================
# Inline Strategy (No Isolation)
# =============================================================================


class InlineExecutionStrategy(ExecutionStrategy):
    """Inline execution without isolation.

    Useful for debugging and trusted operations.
    No timeout protection.
    """

    name = ExecutionBackend.INLINE

    def execute(
        self,
        func: Callable[[], T],
        timeout_seconds: float,
        **kwargs: Any,
    ) -> ExecutionResult[T]:
        """Execute inline without isolation."""
        metrics = ExecutionMetrics(backend_used=self.name)

        try:
            value = func()
            return ExecutionResult.ok(value, metrics)
        except Exception as e:
            return ExecutionResult.failure(e, metrics)

    def is_available(self) -> bool:
        return True


# =============================================================================
# Strategy Registry
# =============================================================================


class ExecutionStrategyRegistry:
    """Registry for execution strategies."""

    def __init__(self) -> None:
        self._strategies: dict[ExecutionBackend, ExecutionStrategy] = {}
        self._register_defaults()

    def _register_defaults(self) -> None:
        """Register default strategies."""
        self.register(ThreadExecutionStrategy())
        self.register(ProcessExecutionStrategy())
        self.register(AdaptiveExecutionStrategy())
        self.register(InlineExecutionStrategy())

    def register(self, strategy: ExecutionStrategy) -> None:
        """Register a strategy."""
        self._strategies[strategy.name] = strategy

    def get(self, backend: ExecutionBackend) -> ExecutionStrategy:
        """Get strategy by backend type."""
        if backend not in self._strategies:
            raise KeyError(f"Unknown backend: {backend}")
        return self._strategies[backend]

    def get_available(self) -> list[ExecutionBackend]:
        """Get list of available backends."""
        return [
            backend
            for backend, strategy in self._strategies.items()
            if strategy.is_available()
        ]


# Global registry
execution_strategy_registry = ExecutionStrategyRegistry()


# =============================================================================
# Resource Monitor
# =============================================================================


@dataclass
class ResourceLimits:
    """Resource limits for execution.

    Attributes:
        max_memory_mb: Maximum memory usage
        max_cpu_percent: Maximum CPU usage
        max_open_files: Maximum open file descriptors
    """

    max_memory_mb: float = 0  # 0 = unlimited
    max_cpu_percent: float = 0  # 0 = unlimited
    max_open_files: int = 0  # 0 = unlimited


@dataclass
class ResourceUsage:
    """Current resource usage."""

    memory_mb: float = 0.0
    cpu_percent: float = 0.0
    open_files: int = 0


class ResourceMonitor:
    """Monitors resource usage during execution.

    Uses psutil if available, otherwise provides estimates.
    """

    def __init__(self):
        self._psutil_available = self._check_psutil()

    def _check_psutil(self) -> bool:
        """Check if psutil is available."""
        try:
            import psutil
            return True
        except ImportError:
            return False

    def get_current_usage(self) -> ResourceUsage:
        """Get current resource usage."""
        if not self._psutil_available:
            return ResourceUsage()

        try:
            import psutil
            process = psutil.Process()

            memory_info = process.memory_info()
            memory_mb = memory_info.rss / (1024 * 1024)

            cpu_percent = process.cpu_percent()

            try:
                open_files = len(process.open_files())
            except Exception:
                open_files = 0

            return ResourceUsage(
                memory_mb=memory_mb,
                cpu_percent=cpu_percent,
                open_files=open_files,
            )
        except Exception:
            return ResourceUsage()

    def exceeds_limits(
        self,
        usage: ResourceUsage,
        limits: ResourceLimits,
    ) -> tuple[bool, str]:
        """Check if usage exceeds limits.

        Returns:
            Tuple of (exceeds, reason)
        """
        if limits.max_memory_mb > 0 and usage.memory_mb > limits.max_memory_mb:
            return True, f"Memory usage {usage.memory_mb:.1f}MB exceeds limit {limits.max_memory_mb:.1f}MB"

        if limits.max_cpu_percent > 0 and usage.cpu_percent > limits.max_cpu_percent:
            return True, f"CPU usage {usage.cpu_percent:.1f}% exceeds limit {limits.max_cpu_percent:.1f}%"

        if limits.max_open_files > 0 and usage.open_files > limits.max_open_files:
            return True, f"Open files {usage.open_files} exceeds limit {limits.max_open_files}"

        return False, ""


# Global monitor
resource_monitor = ResourceMonitor()


# =============================================================================
# Process Timeout Executor (Main Interface)
# =============================================================================


@dataclass
class ProcessTimeoutConfig:
    """Configuration for process timeout executor.

    Attributes:
        default_timeout_seconds: Default timeout
        default_backend: Default execution backend
        enable_circuit_breaker: Use circuit breaker
        enable_complexity_estimation: Pre-check complexity
        enable_resource_monitoring: Monitor resources
        resource_limits: Resource limits
        graceful_termination_seconds: Time before force kill
        max_retries: Maximum retry attempts
        retry_backoff_factor: Backoff multiplier for retries
    """

    default_timeout_seconds: float = 60.0
    default_backend: ExecutionBackend = ExecutionBackend.ADAPTIVE
    enable_circuit_breaker: bool = True
    enable_complexity_estimation: bool = True
    enable_resource_monitoring: bool = True
    resource_limits: ResourceLimits = field(default_factory=ResourceLimits)
    graceful_termination_seconds: float = 2.0
    max_retries: int = 0
    retry_backoff_factor: float = 2.0

    @classmethod
    def strict(cls) -> "ProcessTimeoutConfig":
        """Create strict configuration."""
        return cls(
            default_timeout_seconds=30.0,
            enable_circuit_breaker=True,
            resource_limits=ResourceLimits(max_memory_mb=1000),
            max_retries=0,
        )

    @classmethod
    def lenient(cls) -> "ProcessTimeoutConfig":
        """Create lenient configuration."""
        return cls(
            default_timeout_seconds=300.0,
            enable_circuit_breaker=False,
            max_retries=2,
        )

    @classmethod
    def fast(cls) -> "ProcessTimeoutConfig":
        """Create fast configuration (thread-based)."""
        return cls(
            default_timeout_seconds=10.0,
            default_backend=ExecutionBackend.THREAD,
            enable_complexity_estimation=False,
        )

    @classmethod
    def safe(cls) -> "ProcessTimeoutConfig":
        """Create safe configuration (process-based)."""
        return cls(
            default_timeout_seconds=60.0,
            default_backend=ExecutionBackend.PROCESS,
            enable_circuit_breaker=True,
            enable_complexity_estimation=True,
        )


class ProcessTimeoutExecutor:
    """Enterprise-grade timeout executor with process isolation.

    This is the main interface for executing operations with reliable
    timeout control, including native code like Polars.

    Features:
    - Process isolation for reliable termination
    - Pre-execution complexity estimation
    - Circuit breaker for cascade prevention
    - Resource monitoring
    - Retry with backoff

    Example:
        # Basic usage
        executor = ProcessTimeoutExecutor()
        result = executor.execute(
            lambda: expensive_operation(),
            timeout_seconds=30,
        )

        if result.success:
            print(result.value)
        elif result.timed_out:
            print("Operation timed out")
        else:
            print(f"Error: {result.error}")

        # With hints for better execution
        result = executor.execute_with_hints(
            lambda: profile_column(df, "email"),
            timeout_seconds=60,
            operation_type="profile_column",
            data_size=1_000_000,
        )
    """

    def __init__(
        self,
        config: ProcessTimeoutConfig | None = None,
        estimator: ComplexityEstimator | None = None,
    ):
        """Initialize executor.

        Args:
            config: Executor configuration
            estimator: Complexity estimator
        """
        self.config = config or ProcessTimeoutConfig()
        self.estimator = estimator or default_complexity_estimator
        self._circuit_breakers: dict[str, CircuitBreaker] = {}
        self._lock = threading.RLock()

    def execute(
        self,
        func: Callable[[], T],
        timeout_seconds: float | None = None,
        backend: ExecutionBackend | None = None,
        operation_name: str = "operation",
    ) -> ExecutionResult[T]:
        """Execute function with timeout.

        Args:
            func: Function to execute
            timeout_seconds: Timeout (uses config default if None)
            backend: Execution backend (uses config default if None)
            operation_name: Name for logging and circuit breaker

        Returns:
            Execution result
        """
        timeout = timeout_seconds or self.config.default_timeout_seconds
        backend = backend or self.config.default_backend

        # Check circuit breaker
        if self.config.enable_circuit_breaker:
            breaker = self._get_circuit_breaker(operation_name)
            if not breaker.can_execute():
                return ExecutionResult.failure(
                    RuntimeError(f"Circuit breaker open for '{operation_name}'"),
                    ExecutionMetrics(),
                )

        # Check resources
        if self.config.enable_resource_monitoring:
            usage = resource_monitor.get_current_usage()
            exceeds, reason = resource_monitor.exceeds_limits(
                usage, self.config.resource_limits
            )
            if exceeds:
                return ExecutionResult.failure(
                    RuntimeError(f"Resource limit exceeded: {reason}"),
                    ExecutionMetrics(),
                )

        # Execute with retry
        result = self._execute_with_retry(func, timeout, backend, operation_name)

        # Update circuit breaker
        if self.config.enable_circuit_breaker:
            if result.success:
                breaker.record_success()
            else:
                breaker.record_failure()

        return result

    def execute_with_hints(
        self,
        func: Callable[[], T],
        timeout_seconds: float | None = None,
        operation_type: str = "default",
        data_size: int = 0,
        column_count: int = 1,
    ) -> ExecutionResult[T]:
        """Execute with complexity hints for better decisions.

        Args:
            func: Function to execute
            timeout_seconds: Timeout
            operation_type: Type of operation
            data_size: Number of rows
            column_count: Number of columns

        Returns:
            Execution result
        """
        timeout = timeout_seconds or self.config.default_timeout_seconds

        # Estimate complexity
        if self.config.enable_complexity_estimation:
            estimate = self.estimator.estimate(
                operation_type=operation_type,
                data_size=data_size,
                column_count=column_count,
            )

            # Check if operation will likely timeout
            if estimate.exceeds_timeout(timeout):
                logger.warning(
                    f"Operation '{operation_type}' estimated to take "
                    f"{estimate.estimated_time_seconds:.1f}s, exceeds timeout {timeout}s. "
                    f"Consider sampling to {estimate.max_safe_rows} rows."
                )

            # Use recommended backend
            backend = estimate.recommendation
        else:
            backend = self.config.default_backend

        return self.execute(
            func,
            timeout_seconds=timeout,
            backend=backend,
            operation_name=operation_type,
        )

    def execute_safe(
        self,
        func: Callable[[], T],
        timeout_seconds: float | None = None,
        default: T | None = None,
    ) -> T | None:
        """Execute and return default on failure.

        Args:
            func: Function to execute
            timeout_seconds: Timeout
            default: Default value on failure

        Returns:
            Result value or default
        """
        result = self.execute(func, timeout_seconds)
        if result.success:
            return result.value
        return default

    def _execute_with_retry(
        self,
        func: Callable[[], T],
        timeout: float,
        backend: ExecutionBackend,
        operation_name: str,
    ) -> ExecutionResult[T]:
        """Execute with retry logic."""
        strategy = execution_strategy_registry.get(backend)
        retries = 0
        current_timeout = timeout

        while True:
            result = strategy.execute(func, current_timeout)
            result.metrics.retries = retries

            if result.success or not result.timed_out:
                return result

            # Retry logic
            if retries >= self.config.max_retries:
                return result

            retries += 1
            current_timeout *= self.config.retry_backoff_factor

            logger.info(
                f"Retrying '{operation_name}' (attempt {retries + 1}/{self.config.max_retries + 1}), "
                f"timeout={current_timeout:.1f}s"
            )

    def _get_circuit_breaker(self, name: str) -> CircuitBreaker:
        """Get or create circuit breaker for operation."""
        with self._lock:
            if name not in self._circuit_breakers:
                self._circuit_breakers[name] = CircuitBreaker(name)
            return self._circuit_breakers[name]

    def get_stats(self) -> dict[str, Any]:
        """Get executor statistics."""
        return {
            "config": {
                "default_timeout": self.config.default_timeout_seconds,
                "default_backend": self.config.default_backend.value,
            },
            "circuit_breakers": {
                name: breaker.get_stats()
                for name, breaker in self._circuit_breakers.items()
            },
        }

    def reset_circuit_breakers(self) -> None:
        """Reset all circuit breakers."""
        with self._lock:
            for breaker in self._circuit_breakers.values():
                breaker.reset()


# =============================================================================
# Convenience Functions
# =============================================================================


def with_process_timeout(
    func: Callable[[], T],
    timeout_seconds: float,
    default: T | None = None,
) -> T | None:
    """Execute function with process-based timeout.

    Simple convenience function for one-off executions.

    Args:
        func: Function to execute
        timeout_seconds: Timeout in seconds
        default: Value to return on timeout/failure

    Returns:
        Function result or default

    Example:
        result = with_process_timeout(
            lambda: expensive_polars_operation(),
            timeout_seconds=30,
            default=None,
        )
    """
    executor = ProcessTimeoutExecutor()
    return executor.execute_safe(func, timeout_seconds, default)


def estimate_execution_time(
    operation_type: str,
    data_size: int,
    column_count: int = 1,
) -> ComplexityEstimate:
    """Estimate execution time for an operation.

    Args:
        operation_type: Type of operation
        data_size: Number of rows
        column_count: Number of columns

    Returns:
        Complexity estimate

    Example:
        estimate = estimate_execution_time("pattern_match", 1_000_000)
        print(f"Estimated time: {estimate.estimated_time_seconds:.1f}s")
        print(f"Recommended backend: {estimate.recommendation.value}")
    """
    return default_complexity_estimator.estimate(
        operation_type=operation_type,
        data_size=data_size,
        column_count=column_count,
    )


def create_timeout_executor(
    timeout_seconds: float = 60.0,
    backend: str = "adaptive",
    enable_circuit_breaker: bool = True,
) -> ProcessTimeoutExecutor:
    """Create a configured timeout executor.

    Args:
        timeout_seconds: Default timeout
        backend: Execution backend (thread, process, adaptive)
        enable_circuit_breaker: Enable circuit breaker

    Returns:
        Configured executor

    Example:
        executor = create_timeout_executor(
            timeout_seconds=30,
            backend="process",
        )
    """
    config = ProcessTimeoutConfig(
        default_timeout_seconds=timeout_seconds,
        default_backend=ExecutionBackend(backend),
        enable_circuit_breaker=enable_circuit_breaker,
    )
    return ProcessTimeoutExecutor(config)


# =============================================================================
# Context Manager
# =============================================================================


@contextmanager
def process_timeout_context(
    timeout_seconds: float,
    operation_name: str = "operation",
) -> Generator[ProcessTimeoutExecutor, None, None]:
    """Context manager for process timeout execution.

    Args:
        timeout_seconds: Timeout in seconds
        operation_name: Name for logging

    Yields:
        Executor instance

    Example:
        with process_timeout_context(30.0, "profiling") as executor:
            result = executor.execute(lambda: profile(data))
    """
    config = ProcessTimeoutConfig(default_timeout_seconds=timeout_seconds)
    executor = ProcessTimeoutExecutor(config)

    try:
        yield executor
    finally:
        # Cleanup if needed
        pass


# =============================================================================
# Decorator
# =============================================================================


def timeout_protected(
    timeout_seconds: float = 60.0,
    backend: ExecutionBackend = ExecutionBackend.ADAPTIVE,
    default: Any = None,
) -> Callable[[Callable[..., T]], Callable[..., T | None]]:
    """Decorator to add timeout protection to a function.

    Args:
        timeout_seconds: Timeout in seconds
        backend: Execution backend
        default: Default value on timeout

    Returns:
        Decorated function

    Example:
        @timeout_protected(timeout_seconds=30)
        def expensive_operation(data):
            return process(data)

        result = expensive_operation(my_data)  # Will timeout after 30s
    """
    def decorator(func: Callable[..., T]) -> Callable[..., T | None]:
        executor = ProcessTimeoutExecutor(
            ProcessTimeoutConfig(
                default_timeout_seconds=timeout_seconds,
                default_backend=backend,
            )
        )

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T | None:
            result = executor.execute(
                lambda: func(*args, **kwargs),
                operation_name=func.__name__,
            )
            if result.success:
                return result.value
            return default

        return wrapper

    return decorator



"""Version resolver for compatibility checking.

This module provides version compatibility checking and resolution
for plugins and their dependencies.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any

from truthound.plugins.versioning.constraints import VersionConstraint

logger = logging.getLogger(__name__)


class CompatibilityLevel(Enum):
    """Level of compatibility between versions."""

    COMPATIBLE = auto()      # Fully compatible
    COMPATIBLE_WARNING = auto()  # Compatible with warnings
    INCOMPATIBLE = auto()    # Not compatible
    UNKNOWN = auto()         # Cannot determine


@dataclass(frozen=True)
class CompatibilityReport:
    """Report on version compatibility.

    Provides detailed information about compatibility between
    a plugin version and host version.

    Attributes:
        is_compatible: Whether versions are compatible
        level: Compatibility level
        host_version: Host application version
        plugin_version: Plugin version
        required_constraint: Version constraint from plugin
        errors: List of compatibility errors
        warnings: List of compatibility warnings
        migration_hints: Suggested migration steps
    """

    is_compatible: bool
    level: CompatibilityLevel
    host_version: str
    plugin_version: str
    required_constraint: VersionConstraint
    errors: tuple[str, ...] = ()
    warnings: tuple[str, ...] = ()
    migration_hints: tuple[str, ...] = ()

    @classmethod
    def compatible(
        cls,
        host_version: str,
        plugin_version: str,
        constraint: VersionConstraint,
        warnings: tuple[str, ...] = (),
    ) -> "CompatibilityReport":
        """Create a compatible report."""
        level = CompatibilityLevel.COMPATIBLE_WARNING if warnings else CompatibilityLevel.COMPATIBLE
        return cls(
            is_compatible=True,
            level=level,
            host_version=host_version,
            plugin_version=plugin_version,
            required_constraint=constraint,
            warnings=warnings,
        )

    @classmethod
    def incompatible(
        cls,
        host_version: str,
        plugin_version: str,
        constraint: VersionConstraint,
        errors: tuple[str, ...],
        migration_hints: tuple[str, ...] = (),
    ) -> "CompatibilityReport":
        """Create an incompatible report."""
        return cls(
            is_compatible=False,
            level=CompatibilityLevel.INCOMPATIBLE,
            host_version=host_version,
            plugin_version=plugin_version,
            required_constraint=constraint,
            errors=errors,
            migration_hints=migration_hints,
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            "is_compatible": self.is_compatible,
            "level": self.level.name,
            "host_version": self.host_version,
            "plugin_version": self.plugin_version,
            "constraint": str(self.required_constraint),
            "errors": list(self.errors),
            "warnings": list(self.warnings),
            "migration_hints": list(self.migration_hints),
        }


class VersionResolver:
    """Resolves version compatibility for plugins.

    Checks if plugin versions are compatible with the host version
    and provides detailed compatibility reports.

    Example:
        >>> resolver = VersionResolver()
        >>> constraint = VersionConstraint(min_version="0.1.0")
        >>> report = resolver.check_compatibility("0.5.0", constraint, "0.2.0")
        >>> print(report.is_compatible)
    """

    def __init__(self) -> None:
        """Initialize version resolver."""
        self._breaking_changes: dict[str, list[str]] = {}

    def register_breaking_change(
        self,
        version: str,
        description: str,
    ) -> None:
        """Register a breaking change at a version.

        Args:
            version: Version where breaking change occurred
            description: Description of the breaking change
        """
        if version not in self._breaking_changes:
            self._breaking_changes[version] = []
        self._breaking_changes[version].append(description)

    def check_compatibility(
        self,
        plugin_version: str,
        constraint: VersionConstraint,
        host_version: str,
    ) -> CompatibilityReport:
        """Check if plugin is compatible with host version.

        Args:
            plugin_version: Version of the plugin
            constraint: Version constraint from plugin manifest
            host_version: Current host application version

        Returns:
            CompatibilityReport with detailed information
        """
        errors: list[str] = []
        warnings: list[str] = []
        hints: list[str] = []

        # Check if host version satisfies plugin's constraint
        if not constraint.is_satisfied_by(host_version):
            if constraint.min_version:
                errors.append(
                    f"Plugin requires host version >= {constraint.min_version}, "
                    f"but current version is {host_version}"
                )
                hints.append(
                    f"Upgrade host to at least version {constraint.min_version}"
                )

            if constraint.max_version:
                max_ver = constraint.max_version.rstrip("*")
                errors.append(
                    f"Plugin requires host version < {max_ver}, "
                    f"but current version is {host_version}"
                )
                hints.append(
                    f"Use a plugin version compatible with host {host_version}"
                )

            if host_version in constraint.excluded_versions:
                errors.append(
                    f"Plugin explicitly excludes host version {host_version}"
                )

        # Check for breaking changes
        breaking = self._check_breaking_changes(
            plugin_version,
            host_version,
        )
        if breaking:
            for change in breaking:
                warnings.append(f"Breaking change: {change}")
            hints.extend([
                "Review the changelog for breaking changes",
                "Update plugin code for compatibility",
            ])

        if errors:
            return CompatibilityReport.incompatible(
                host_version=host_version,
                plugin_version=plugin_version,
                constraint=constraint,
                errors=tuple(errors),
                migration_hints=tuple(hints),
            )

        return CompatibilityReport.compatible(
            host_version=host_version,
            plugin_version=plugin_version,
            constraint=constraint,
            warnings=tuple(warnings),
        )

    def _check_breaking_changes(
        self,
        plugin_version: str,
        host_version: str,
    ) -> list[str]:
        """Check for breaking changes between versions."""
        changes = []

        # Get breaking changes in host versions since plugin was written
        for version, descriptions in self._breaking_changes.items():
            # Check if this version is between plugin and host
            if self._is_version_between(version, plugin_version, host_version):
                changes.extend(descriptions)

        return changes

    def _is_version_between(
        self,
        version: str,
        low: str,
        high: str,
    ) -> bool:
        """Check if version is between low and high."""
        try:
            from packaging.version import Version
            v = Version(version)
            l = Version(low)
            h = Version(high)
            return l <= v <= h
        except ImportError:
            # Simple comparison
            return low <= version <= high

    def find_compatible_versions(
        self,
        constraint: VersionConstraint,
        available_versions: list[str],
    ) -> list[str]:
        """Find all versions that satisfy a constraint.

        Args:
            constraint: Version constraint to satisfy
            available_versions: List of available versions

        Returns:
            List of compatible versions (sorted newest first)
        """
        compatible = [
            v for v in available_versions
            if constraint.is_satisfied_by(v)
        ]

        # Sort by version (newest first)
        try:
            from packaging.version import Version
            compatible.sort(key=Version, reverse=True)
        except ImportError:
            compatible.sort(reverse=True)

        return compatible

    def find_best_version(
        self,
        constraint: VersionConstraint,
        available_versions: list[str],
        prefer_stable: bool = True,
    ) -> str | None:
        """Find the best version that satisfies a constraint.

        Args:
            constraint: Version constraint to satisfy
            available_versions: List of available versions
            prefer_stable: Prefer stable over pre-release

        Returns:
            Best matching version or None
        """
        compatible = self.find_compatible_versions(constraint, available_versions)

        if not compatible:
            return None

        if prefer_stable:
            # Filter out pre-releases
            try:
                from packaging.version import Version
                stable = [v for v in compatible if not Version(v).is_prerelease]
                if stable:
                    return stable[0]
            except ImportError:
                pass

        return compatible[0]

    def check_upgrade_path(
        self,
        from_version: str,
        to_version: str,
    ) -> list[str]:
        """Get list of breaking changes in upgrade path.

        Args:
            from_version: Starting version
            to_version: Target version

        Returns:
            List of breaking change descriptions
        """
        changes = []

        for version, descriptions in self._breaking_changes.items():
            if self._is_version_between(version, from_version, to_version):
                changes.extend(descriptions)

        return changes



"""Integration tests for Kafka adapter using Testcontainers.

These tests require Docker to be running and testcontainers installed.
Run with: pytest tests/integration/realtime/test_kafka_adapter.py -v
"""

from __future__ import annotations

import asyncio
import pytest
from datetime import datetime, timezone

# Skip all tests if dependencies not available
pytest_plugins = ["pytest_asyncio"]

try:
    import testcontainers
    TESTCONTAINERS_AVAILABLE = True
except ImportError:
    TESTCONTAINERS_AVAILABLE = False

try:
    import aiokafka
    AIOKAFKA_AVAILABLE = True
except ImportError:
    AIOKAFKA_AVAILABLE = False


@pytest.fixture(scope="module")
def event_loop():
    """Create event loop for async tests."""
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.mark.skipif(
    not TESTCONTAINERS_AVAILABLE,
    reason="testcontainers not installed"
)
@pytest.mark.skipif(
    not AIOKAFKA_AVAILABLE,
    reason="aiokafka not installed"
)
class TestKafkaAdapterIntegration:
    """Integration tests for KafkaAdapter with real Kafka."""

    @pytest.fixture(scope="class")
    async def kafka_container(self):
        """Start Kafka container for tests."""
        from truthound.realtime.testing.containers import KafkaTestContainer

        async with KafkaTestContainer() as kafka:
            # Create test topic
            await kafka.create_topic("test-topic", num_partitions=3)
            yield kafka

    @pytest.fixture
    def kafka_config(self, kafka_container):
        """Get Kafka adapter configuration."""
        from truthound.realtime.adapters.kafka import KafkaAdapterConfig

        return KafkaAdapterConfig(
            bootstrap_servers=kafka_container.get_connection_url(),
            topic="test-topic",
            consumer_group="test-consumer",
            enable_auto_commit=False,
        )

    @pytest.mark.asyncio
    async def test_connect_and_disconnect(self, kafka_config):
        """Test connecting and disconnecting from Kafka."""
        from truthound.realtime.adapters.kafka import KafkaAdapter

        adapter = KafkaAdapter(kafka_config)

        # Connect
        await adapter.connect()
        assert adapter.is_connected

        # Disconnect
        await adapter.disconnect()
        assert not adapter.is_connected

    @pytest.mark.asyncio
    async def test_produce_and_consume(self, kafka_config):
        """Test producing and consuming messages."""
        from truthound.realtime.adapters.kafka import KafkaAdapter
        from truthound.realtime.protocols import StreamMessage

        adapter = KafkaAdapter(kafka_config)

        async with adapter:
            # Produce messages
            for i in range(10):
                message = StreamMessage(
                    key=f"key-{i}",
                    value={"id": i, "data": f"test-{i}"},
                    partition=-1,
                    offset=-1,
                    timestamp=datetime.now(timezone.utc),
                    headers=(),
                    topic="test-topic",
                )
                await adapter.produce(message)

            await adapter.flush()

            # Consume messages
            consumed = []
            timeout = asyncio.get_event_loop().time() + 10  # 10 second timeout

            async for msg in adapter.consume():
                consumed.append(msg)
                await adapter.commit(msg)

                if len(consumed) >= 10:
                    break
                if asyncio.get_event_loop().time() > timeout:
                    break

            assert len(consumed) == 10

    @pytest.mark.asyncio
    async def test_batch_produce(self, kafka_config):
        """Test batch message production."""
        from truthound.realtime.adapters.kafka import KafkaAdapter
        from truthound.realtime.protocols import StreamMessage

        adapter = KafkaAdapter(kafka_config)

        async with adapter:
            messages = [
                StreamMessage(
                    key=f"batch-key-{i}",
                    value={"batch_id": i},
                    partition=-1,
                    offset=-1,
                    timestamp=datetime.now(timezone.utc),
                    headers=(),
                    topic="test-topic",
                )
                for i in range(100)
            ]

            await adapter.produce_batch(messages)
            await adapter.flush()

            # Verify metrics
            metrics = adapter.metrics
            assert metrics.messages_produced >= 100

    @pytest.mark.asyncio
    async def test_partition_info(self, kafka_config, kafka_container):
        """Test getting partition information."""
        from truthound.realtime.adapters.kafka import KafkaAdapter

        adapter = KafkaAdapter(kafka_config)

        async with adapter:
            partitions = await adapter.get_partition_info()
            assert "test-topic" in partitions
            assert len(partitions["test-topic"]) == 3  # We created 3 partitions

    @pytest.mark.asyncio
    async def test_consumer_lag(self, kafka_config):
        """Test consumer lag calculation."""
        from truthound.realtime.adapters.kafka import KafkaAdapter

        adapter = KafkaAdapter(kafka_config)

        async with adapter:
            lag = await adapter.get_consumer_lag()
            assert "test-topic" in lag
            # Lag should be a dict of partition -> lag value
            assert all(isinstance(v, int) for v in lag["test-topic"].values())

    @pytest.mark.asyncio
    async def test_reconnection(self, kafka_config):
        """Test adapter reconnection."""
        from truthound.realtime.adapters.kafka import KafkaAdapter

        adapter = KafkaAdapter(kafka_config)

        await adapter.connect()
        assert adapter.is_connected

        # Simulate reconnection
        await adapter.reconnect()
        assert adapter.is_connected

        await adapter.disconnect()

    @pytest.mark.asyncio
    async def test_error_handling(self, kafka_config):
        """Test error handling for invalid operations."""
        from truthound.realtime.adapters.kafka import KafkaAdapter
        from truthound.realtime.protocols import StreamMessage, ConnectionError

        adapter = KafkaAdapter(kafka_config)

        # Try to produce without connecting
        message = StreamMessage(
            key="test",
            value={"test": True},
            partition=-1,
            offset=-1,
            timestamp=datetime.now(timezone.utc),
            headers=(),
            topic="test-topic",
        )

        with pytest.raises(ConnectionError):
            await adapter.produce(message)


@pytest.mark.skipif(
    not TESTCONTAINERS_AVAILABLE,
    reason="testcontainers not installed"
)
@pytest.mark.skipif(
    not AIOKAFKA_AVAILABLE,
    reason="aiokafka not installed"
)
class TestKafkaStreamingPipeline:
    """Integration tests for complete streaming pipeline with Kafka."""

    @pytest.fixture(scope="class")
    async def kafka_env(self):
        """Start Kafka environment for pipeline tests."""
        from truthound.realtime.testing.containers import StreamTestEnvironment

        async with StreamTestEnvironment(["kafka"]) as env:
            await env.create_kafka_topic("pipeline-topic", num_partitions=3)
            yield env

    @pytest.mark.asyncio
    async def test_validation_pipeline(self, kafka_env):
        """Test complete validation pipeline with Kafka."""
        from truthound.realtime.adapters.kafka import KafkaAdapter, KafkaAdapterConfig
        from truthound.realtime.protocols import StreamMessage
        from truthound.realtime.testing.fixtures import TestFixtures

        config = KafkaAdapterConfig(
            bootstrap_servers=kafka_env.get_service_url("kafka"),
            topic="pipeline-topic",
            consumer_group="pipeline-consumer",
        )

        adapter = KafkaAdapter(config)
        fixtures = TestFixtures()

        async with adapter:
            # Produce test messages
            test_data = fixtures.default_messages(50)
            for data in test_data:
                message = StreamMessage(
                    key=str(data.get("id", "unknown")),
                    value=data,
                    partition=-1,
                    offset=-1,
                    timestamp=datetime.now(timezone.utc),
                    headers=(),
                    topic="pipeline-topic",
                )
                await adapter.produce(message)

            await adapter.flush()

            # Consume and validate
            valid_count = 0
            invalid_count = 0
            consumed = 0

            async for msg in adapter.consume():
                consumed += 1
                value = msg.value

                # Simple validation
                if value.get("id") is not None and value.get("value") is not None:
                    valid_count += 1
                else:
                    invalid_count += 1

                await adapter.commit(msg)

                if consumed >= 50:
                    break

            assert consumed == 50
            assert valid_count == 50  # All default messages should be valid

    @pytest.mark.asyncio
    async def test_window_processing(self, kafka_env):
        """Test window-based processing with Kafka."""
        from truthound.realtime.adapters.kafka import KafkaAdapter, KafkaAdapterConfig
        from truthound.realtime.protocols import StreamMessage
        from truthound.realtime.processing.windows import (
            TumblingWindowProcessor,
            WindowConfig,
            SumAggregator,
        )

        config = KafkaAdapterConfig(
            bootstrap_servers=kafka_env.get_service_url("kafka"),
            topic="pipeline-topic",
            consumer_group="window-consumer",
        )

        adapter = KafkaAdapter(config)

        # Create window processor
        window_config = WindowConfig(window_size_seconds=5)
        processor = TumblingWindowProcessor(
            config=window_config,
            aggregators={"total": SumAggregator("value")},
        )

        async with adapter:
            # Produce messages with values
            for i in range(20):
                message = StreamMessage(
                    key=f"window-{i}",
                    value={"value": 10},  # Each message has value 10
                    partition=-1,
                    offset=-1,
                    timestamp=datetime.now(timezone.utc),
                    headers=(),
                    topic="pipeline-topic",
                )
                await adapter.produce(message)

            await adapter.flush()

            # Consume and process through window
            consumed = 0
            async for msg in adapter.consume():
                processor.add(msg)
                consumed += 1

                if consumed >= 20:
                    break

            # Check aggregated results
            results = processor.get_results()
            # Sum of 20 messages * 10 value each = 200
            assert results.get("total", 0) >= 100  # At least some aggregated


@pytest.mark.skipif(
    not TESTCONTAINERS_AVAILABLE,
    reason="testcontainers not installed"
)
class TestMultiContainerEnvironment:
    """Tests for multi-container test environment."""

    @pytest.mark.asyncio
    async def test_kafka_and_redis(self):
        """Test environment with Kafka and Redis."""
        from truthound.realtime.testing.containers import StreamTestEnvironment

        async with StreamTestEnvironment(["kafka", "redis"]) as env:
            kafka_url = env.get_service_url("kafka")
            redis_url = env.get_service_url("redis")

            assert ":" in kafka_url  # Should have host:port
            assert "redis://" in redis_url

            # Create topic
            await env.create_kafka_topic("multi-test", num_partitions=1)

    @pytest.mark.asyncio
    async def test_container_lifecycle(self):
        """Test container start/stop lifecycle."""
        from truthound.realtime.testing.containers import (
            KafkaTestContainer,
            RedisTestContainer,
        )

        kafka = KafkaTestContainer()
        redis = RedisTestContainer()

        # Start containers
        await kafka.start()
        await redis.start()

        assert kafka.is_running
        assert redis.is_running

        # Get URLs
        kafka_url = kafka.get_connection_url()
        redis_url = redis.get_connection_url()

        assert kafka_url is not None
        assert redis_url is not None

        # Stop containers
        await kafka.stop()
        await redis.stop()

        assert not kafka.is_running
        assert not redis.is_running



"""Tests for i18n catalog module."""

import pytest
from truthound.datadocs.i18n.catalog import (
    ReportCatalog,
    CatalogBuilder,
    CatalogRegistry,
    get_catalog,
    get_supported_locales,
)


class TestReportCatalog:
    """Tests for ReportCatalog class."""

    def test_create_empty_catalog(self):
        """Test creating empty catalog."""
        catalog = ReportCatalog(locale="en")
        assert catalog.locale == "en"
        assert len(catalog) == 0

    def test_create_catalog_with_messages(self):
        """Test creating catalog with messages."""
        catalog = ReportCatalog(
            locale="ko",
            messages={"report.title": "데이터 품질 보고서"},
        )
        assert catalog["report.title"] == "데이터 품질 보고서"

    def test_get_message(self):
        """Test getting a message."""
        catalog = ReportCatalog(
            locale="en",
            messages={"key": "value"},
        )
        assert catalog.get("key") == "value"
        assert catalog.get("nonexistent") == "nonexistent"
        assert catalog.get("nonexistent", "default") == "default"

    def test_get_with_params(self):
        """Test getting message with parameters."""
        catalog = ReportCatalog(
            locale="en",
            messages={"alert.count": "{count} alerts found"},
        )
        result = catalog.get("alert.count", count=5)
        assert result == "5 alerts found"

    def test_contains(self):
        """Test __contains__."""
        catalog = ReportCatalog(
            locale="en",
            messages={"key": "value"},
        )
        assert "key" in catalog
        assert "nonexistent" not in catalog

    def test_merge(self):
        """Test merging catalogs."""
        catalog1 = ReportCatalog(
            locale="en",
            messages={"a": "1", "b": "2"},
        )
        catalog2 = ReportCatalog(
            locale="en",
            messages={"b": "3", "c": "4"},
        )
        merged = catalog1.merge(catalog2)
        assert merged["a"] == "1"
        assert merged["b"] == "3"  # Overwritten
        assert merged["c"] == "4"

    def test_extend(self):
        """Test extending catalog."""
        catalog = ReportCatalog(
            locale="en",
            messages={"a": "1"},
        )
        extended = catalog.extend({"b": "2"})
        assert extended["a"] == "1"
        assert extended["b"] == "2"
        assert "b" not in catalog  # Original unchanged

    def test_to_dict(self):
        """Test converting to dict."""
        catalog = ReportCatalog(
            locale="en",
            messages={"key": "value"},
            metadata={"name": "English"},
        )
        d = catalog.to_dict()
        assert d["locale"] == "en"
        assert d["messages"]["key"] == "value"
        assert d["metadata"]["name"] == "English"

    def test_from_dict(self):
        """Test creating from dict."""
        catalog = ReportCatalog.from_dict(
            "ja",
            {"report.title": "レポート"},
            metadata={"name": "Japanese"},
        )
        assert catalog.locale == "ja"
        assert catalog["report.title"] == "レポート"


class TestCatalogBuilder:
    """Tests for CatalogBuilder class."""

    def test_builder_basic(self):
        """Test basic builder usage."""
        catalog = (
            CatalogBuilder("en")
            .add("key1", "value1")
            .add("key2", "value2")
            .build()
        )
        assert catalog.locale == "en"
        assert catalog["key1"] == "value1"
        assert catalog["key2"] == "value2"

    def test_builder_report_section(self):
        """Test adding report section."""
        catalog = (
            CatalogBuilder("ko")
            .add_report_section(
                title="보고서",
                subtitle="부제",
                summary="요약",
                details="상세",
            )
            .build()
        )
        assert catalog["report.title"] == "보고서"
        assert catalog["report.subtitle"] == "부제"

    def test_builder_quality_labels(self):
        """Test adding quality labels."""
        catalog = (
            CatalogBuilder("en")
            .add_quality_labels(
                excellent="Excellent",
                good="Good",
                fair="Fair",
                poor="Poor",
                critical="Critical",
            )
            .build()
        )
        assert catalog["quality.excellent"] == "Excellent"
        assert catalog["quality.critical"] == "Critical"

    def test_builder_with_metadata(self):
        """Test adding metadata."""
        catalog = (
            CatalogBuilder("en")
            .with_metadata(name="English", complete=True)
            .build()
        )
        assert catalog.metadata["name"] == "English"
        assert catalog.metadata["complete"] is True


class TestCatalogRegistry:
    """Tests for CatalogRegistry class."""

    def test_get_default_locale(self):
        """Test getting default locale."""
        registry = CatalogRegistry()
        catalog = registry.get("en")
        assert catalog.locale == "en"
        assert "report.title" in catalog

    def test_get_korean(self):
        """Test getting Korean catalog."""
        registry = CatalogRegistry()
        catalog = registry.get("ko")
        assert catalog.locale == "ko"
        assert catalog["report.title"] == "데이터 품질 보고서"

    def test_get_japanese(self):
        """Test getting Japanese catalog."""
        registry = CatalogRegistry()
        catalog = registry.get("ja")
        assert catalog.locale == "ja"
        assert "レポート" in catalog["report.title"]

    def test_get_fallback(self):
        """Test fallback to default locale."""
        registry = CatalogRegistry()
        catalog = registry.get("nonexistent")
        assert catalog.locale == "en"  # Fallback to English

    def test_get_with_region(self):
        """Test getting with region code."""
        registry = CatalogRegistry()
        catalog = registry.get("en_US")
        assert catalog.locale == "en"  # Falls back to base language

    def test_register_custom(self):
        """Test registering custom catalog."""
        registry = CatalogRegistry()
        custom = ReportCatalog(
            locale="custom",
            messages={"test": "Custom message"},
        )
        registry.register(custom)
        result = registry.get("custom")
        assert result["test"] == "Custom message"

    def test_list_locales(self):
        """Test listing locales."""
        registry = CatalogRegistry()
        locales = registry.list_locales()
        assert "en" in locales
        assert "ko" in locales
        assert "ja" in locales


class TestModuleFunctions:
    """Tests for module-level functions."""

    def test_get_catalog(self):
        """Test get_catalog function."""
        catalog = get_catalog("en")
        assert catalog.locale == "en"

    def test_get_supported_locales(self):
        """Test get_supported_locales function."""
        locales = get_supported_locales()
        assert isinstance(locales, list)
        assert len(locales) > 0
        assert "en" in locales



"""Cross-table row count validators."""

from typing import Any

import polars as pl

from truthound.types import Severity
from truthound.validators.base import ValidationIssue, Validator
from truthound.validators.registry import register_validator


@register_validator
class CrossTableRowCountValidator(Validator):
    """Validates that row counts between two tables match.

    Example:
        # Orders should have same count as order_details
        validator = CrossTableRowCountValidator(
            reference_data=order_details_df,
            reference_name="order_details",
        )
        issues = validator.validate(orders_df.lazy())
    """

    name = "cross_table_row_count"
    category = "cross_table"

    def __init__(
        self,
        reference_data: pl.DataFrame | pl.LazyFrame,
        reference_name: str = "reference",
        tolerance: int = 0,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.reference_name = reference_name
        self.tolerance = tolerance

        # Get reference row count
        if isinstance(reference_data, pl.LazyFrame):
            self.reference_count = reference_data.select(pl.len()).collect().item()
        else:
            self.reference_count = len(reference_data)

    def validate(self, lf: pl.LazyFrame) -> list[ValidationIssue]:
        issues: list[ValidationIssue] = []

        source_count = lf.select(pl.len()).collect().item()
        diff = abs(source_count - self.reference_count)

        if diff > self.tolerance:
            issues.append(
                ValidationIssue(
                    column="_table",
                    issue_type="cross_table_row_count_mismatch",
                    count=diff,
                    severity=Severity.CRITICAL,
                    details=f"Row count mismatch with {self.reference_name}",
                    expected=self.reference_count,
                    actual=source_count,
                )
            )

        return issues


@register_validator
class CrossTableRowCountFactorValidator(Validator):
    """Validates that row count equals reference count times a factor.

    Example:
        # Daily aggregates should be ~30x monthly aggregates
        validator = CrossTableRowCountFactorValidator(
            reference_data=monthly_df,
            factor=30,
            tolerance_ratio=0.1,  # Allow 10% variance
        )
    """

    name = "cross_table_row_count_factor"
    category = "cross_table"

    def __init__(
        self,
        reference_data: pl.DataFrame | pl.LazyFrame,
        factor: float,
        reference_name: str = "reference",
        tolerance_ratio: float = 0.0,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.factor = factor
        self.reference_name = reference_name
        self.tolerance_ratio = tolerance_ratio

        if isinstance(reference_data, pl.LazyFrame):
            self.reference_count = reference_data.select(pl.len()).collect().item()
        else:
            self.reference_count = len(reference_data)

    def validate(self, lf: pl.LazyFrame) -> list[ValidationIssue]:
        issues: list[ValidationIssue] = []

        source_count = lf.select(pl.len()).collect().item()
        expected_count = self.reference_count * self.factor
        tolerance = expected_count * self.tolerance_ratio

        if abs(source_count - expected_count) > tolerance:
            ratio = source_count / expected_count if expected_count > 0 else float("inf")
            issues.append(
                ValidationIssue(
                    column="_table",
                    issue_type="cross_table_row_count_factor_mismatch",
                    count=abs(int(source_count - expected_count)),
                    severity=Severity.HIGH,
                    details=f"Expected {self.reference_name} × {self.factor}",
                    expected=f"{expected_count:.0f} (±{self.tolerance_ratio:.0%})",
                    actual=f"{source_count} (ratio: {ratio:.2f}x)",
                )
            )

        return issues
